# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Backend of the simulation controlling:

- Precision
- Device
- Tensor network ansatz
- Settings for the MPI (if any)
- Tensor class used in the simulation
"""
import importlib
import json
import os
from typing import OrderedDict

from qtealeaves.tensors.tensor import GPU_AVAILABLE

from .mpi_utils import MPISettings

__all__ = ["QCBackend"]

# pylint: disable=too-many-arguments


class QCBackend:
    """
    Backend for the simulation. Contains all the informations about
    which executable you want to run.

    Parameters
    ----------
    precision: str, optional
        Precision of the simulation.
        Select a real precision ONLY if you use ONLY real gates.
        Available:
        - "A": automatic. For the heuristic see `self.resolve_precision`.
        - "Z": double precision complex;
        - "C": single precision complex;
        - "D": double precision real;
        - "S": single precision real.
        Default to "A".
    device: str, optional
        Device of the simulation.
        Available:
        - "A" : automatic. For the heuristic see `self.resolve_device`
        - "cpu": use the cpu
        - "gpu": use the gpu if it is available
        - "cpu+gpu": use the device-mixed mode (if gpu is available)
        Default to "A".
    ansatz : str, optional
        Whether to run the circuit with MPS or TTN tensor network ansatz.
        Default to "MPS".
    mpi_settings : MPISettings | None, optional
        Settings for running the simulation multi-node.
        Default to None, i.e. no MPI.
    tensor_module : str, optional
        Module used to perform the computations. Available:
        - `"numpy"` (Default, uses cupy for the GPU)
        - `"torch"`, uses pytorch;
        - `"tensorflow"`
        - `"jax"`.
        Default to `"numpy"`
    """

    def __init__(
        self,
        precision="A",
        device="cpu",
        ansatz="MPS",
        mpi_settings=None,
        tensor_module="numpy",
    ):
        self._precision = precision.upper()
        self._device = device
        self._ansatz = ansatz.upper()
        self.mpi_settings = MPISettings() if mpi_settings is None else mpi_settings
        self.tensor_module = tensor_module.lower()

        if importlib.util.find_spec(self.tensor_module) is None:
            raise ValueError(f"Module {self.tensor_module} is not installed.")

    def to_dict(self):
        """
        Map the backend to a dictionary.
        """

        dictionary = OrderedDict({})
        mpi = "T" if self.num_procs > 1 else "F"
        gpu = "T" if self.device == "gpu" else "F"
        dictionary["simulation_mode"] = self.precision + mpi + gpu
        dictionary["approach"] = self.mpi_approach

        return dictionary

    def resolve_precision(self, min_fidelity, tol=1e-7):
        """
        Resolve the precision of the simulation.
        Heuristic if `self._precision="A"`.

        Parameters
        ----------
        min_fidelity: float
            Lower bound of the fidelity of the simulation at the moment
        tol: float, optional
            Tolerance after which you switch to single precision.
            Default to 1e-7

        Returns
        -------
        str
            The selected precision
        """
        if self._precision != "A":
            return self.precision

        # The lower bound of the fidelity of our state is
        # below the number of digits of a single
        # precision
        if 1 - min_fidelity > tol:
            return "C"

        return "Z"

    def resolve_device(self, bond_dimension, previous_device, exp_gpu=7):
        """
        Resolve the device if it set on automatic.

        Parameters
        ----------
        bond_dimension : int
            Maximum bond dimension of the system
        previous_device : str
            Device where the system is currently. This is used
            to ensure we do not keep exchanging data back and
            forth.
        exp_gpu : int, optional
            Exponent of the bond dimension after which you switch
            to the gpu, i.e:
            - if chi >= 2**exp_gpu -> use gpu
            - if chi <= 2**(exp_gpu-1) -> use cpu
            Default to 7. (switch at 128)

        Returns
        -------
        str
            Device where to move (or keep) the system
        """
        if self._device != "A":
            return self._device

        if bond_dimension >= 2**exp_gpu and GPU_AVAILABLE:
            return "gpu"
        # The condition on the CPU is here because we want
        # to avoid keep exchanging informations if the bond
        # dimension oscillates between 129 and 120
        if bond_dimension <= 2 ** (exp_gpu - 1):
            return "cpu"

        return previous_device

    @property
    def where_barriers(self):
        """
        This parameter is important only if you want to use MPI parallelization,
        where a barrier is equivalent to a canonization in the MPS simulation.
        Default to -1.
        """
        return self.mpi_settings.where_barriers

    @property
    def precision(self):
        """Precision property"""
        return self._precision

    @property
    def device(self):
        """Device property"""
        return self._device

    @property
    def num_procs(self):
        """Number of processes property"""
        return self.mpi_settings.num_procs

    @property
    def mpi_approach(self):
        """mpi_approach property (inherited from `mpi_settings`)"""
        return self.mpi_settings.mpi_approach.upper()

    @property
    def ansatz(self):
        """ansatz property"""
        return self._ansatz

    @property
    def mpi_command(self):
        """mpi_command property"""
        return self.mpi_settings.mpi_command

    @property
    def identifier(self):
        """Identifier combining all properties."""
        return ":".join(
            [
                self.resolve_precision(1),
                self.resolve_device(1, "cpu"),
                str(self.num_procs),
                self.mpi_approach,
                self.ansatz,
            ]
        )

    def to_json(self, path):
        """
        Write the class as a json on file as
        backend.json in the folder path
        """
        path = os.path.join(path, "backend.json")
        dictionary = OrderedDict()
        dictionary["device"] = self.device
        dictionary["precision"] = self.precision
        dictionary["num_procs"] = self.num_procs
        dictionary["ansatz"] = self.ansatz
        dictionary["mpi_approach"] = self.mpi_approach
        dictionary["mpi_command"] = self.mpi_command

        with open(path, "w") as fhandle:
            json.dump(dictionary, fhandle, indent=4)

    @classmethod
    def from_json(cls, path):
        """
        Initialize the class from a json file called
        "backend.json" in the folder path
        """
        path = os.path.join(path, "backend.json")
        with open(path, "r") as fhandle:
            dictionary = json.load(fhandle)[0]

        return cls(**dictionary)



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import pathlib
import shlex
import subprocess
import unittest


class TestFormatting(unittest.TestCase):
    folders = "qmatchatea", "tests"

    def check_ext_util(self, cmd_call):
        # Function to run the command call. Then it checks whether the command's exit code is 0, which means success
        result = subprocess.run(
            shlex.join(cmd_call), stderr=subprocess.PIPE, shell=True, text=True
        )
        self.assertEqual(result.returncode, 0, result.stderr)

    def test_black(self):
        self.check_ext_util(["black", "--check", *self.folders])

    def test_isort(self):
        self.check_ext_util(["isort", *self.folders, "--check", "-v"])



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Circuit class used for the simulations.

The circuit is defined as a list of layers, and each layer is a list of
**Instructions**, i.e. a list [QCOperation, sites]. This structure ensures
a simplified analysis of which operations could be executed in parallel.
Furthermore, we point out a couple of further important elements of the
class:

- qregisters: the quantum registers are a way of keeping track of
    "temporary" qubits, that can be added and removed at will. With
    the class methods :func:`add_qregister` and :func:`remove_qregister`
    it is possible to apply such operations. Then, when you apply an operation
    you can always specify to which register you are referring, without having
    to worry about the indexing. In particular, qubits in the same register
    are naturally ordered from left to right. So, if you want to apply a not
    gate to the left-most qubit of the register a1 you will have to call
    `qcirc.x(0, 'a1')`, since the left-most has index 0 for the register 'a1';
- cregisters: the classical registers, where you store the information of
    projective mesurements happened previously on the system and their probability.
    So, the position `idx` of the cregister will hold the tuple
    ``(meas_result, meas_prob``. Each operation
    can be conditioned on the value of a classical register. By calling the
    method :func:`inspect_cregister(creg, idx)` where `creg` is the name of your
    classical register, you can have access to the value of the bit at the
    index idx. If idx=None instead, the full register is treated as the binary
    representation of a decimal number, and that decimal number is given by
    the function. Remember to add the classical registers you need beforehand,
    using :func:`add_cregister`: the quantum circuit start without any of them!

To add operations to the circuit there are three main ways:

- the :func:`add`, to append the operation at the end of the circuit
- some methods to apply widely-used operations to the circuit with a simplified
    notation.
- the :func:`insert` method, to insert an operation at any given layer of the circuit.
    However, this method is dangerous and is not recommended.

The :func:`str_repr` is meant to give a way to visualize the circuit, but is not needed
for the functioning of the class, and it is still under development.
"""

# pylint: disable=too-many-lines

from copy import deepcopy

import numpy as np
import scipy.sparse as sp
from qiskit import QuantumCircuit

from qmatchatea.utils.tn_utils import QCOperators

from .observables import QCObservableStep

# Gates
# Operations
from .operations import (
    ClassicalCondition,
    QCAddSite,
    QCCnot,
    QCCp,
    QCCz,
    QCHadamard,
    QCMeasureProjective,
    QCOperation,
    QCPgate,
    QCRemoveSite,
    QCRenormalize,
    QCSadggate,
    QCSgate,
    QCSwap,
    QCTadggate,
    QCTgate,
    QCXpauli,
    QCYpauli,
    QCZpauli,
)

__all__ = ["Qcircuit"]


# pylint: disable-next=too-many-public-methods, too-many-instance-attributes
class Qcircuit:
    """
    Class used to represent a quantum circuit in
    the qmatchatea library.
    Each layer of the circuit is a list containing
    instructions, i.e. lists [QCOperation, sites].

    - When iterating over the class the iterators returns
      the layers.
    - Iterating over the layer returns the instructions,
      with all permutations of indexes already taken care
      of, based on the quantum registers.

    This means that, once you obtain the operation, you
    simply need to apply it on the defined sites of the mps,
    without further complexity.

    However, you should not run it by yourself, it is
    preferred to use the :func:`QcMps.run_from_qcirc()`.


    Parameters
    ----------
    num_sites: int
        Number of sites in the circuit
    local_dim: int, optional
        Local dimension of the single site. Default to 2.
    """

    def __init__(self, num_sites, local_dim=2) -> None:
        self._num_sites = num_sites
        self._local_dim = local_dim

        # Registers are a way of keeping track of ancilla sites,
        # i.e. sites that get added and removed in runtime.
        # The structure is a name connected to a np.ndarray. The
        # ndarray has as values the true position of the flattened
        # chain. Sites in a register are always ordered in
        # increasing order from left to right
        self._qregisters = {"default": np.arange(num_sites)}

        # Classical registers keep tracks of the result of projective
        # measurement and their probability along the simulation.
        # They are represented as a two-dimensional vector, with the
        # row being the index of the register and the column either
        # the measurement outcome (index=0) or their probability (index=1)
        # They are all initialized to 0, and when retrieved through
        # the cregister function returns the decimal value of their
        # binary representation
        self._cregisters = {}
        self._cl_occupation_map = {}

        # Circuit list. The outer list contains the full circuit,
        # and then each list is a layer. If no operation is applied
        # on a qubit a "wait" operation is applied. To take into
        # account multiple gates and make the wait management easier
        # we encode a logical ndarray where True means occupied and
        # False empty (wait)
        self._data = []
        self._occupation_map = []
        self._add_layer()

        # Number of one-site operations in the circuit
        self._num_one_site_operations = 0
        # Number of two-site operations in the circuit
        self._num_two_site_operations = 0

        # Used in the writing operation for the 1-1 mapping
        self._namelist = []

    def __len__(self):
        """
        Provide number of operations in the circuit
        """
        return self._num_one_site_operations + self._num_two_site_operations

    def __getitem__(self, key):
        """Overwrite the call for lists, you can access the circuit layers using

        .. code-block::
            circ[0]
            >>>  [QCOperation, [0]], [QCOperation, [1, 2]]

        Parameters
        ----------
        key : int
            index of the circuit layer you are interested in

        Returns
        -------
        list
            list of instructions [Operation, sites]
        """
        return self._data[key]

    def __setitem__(self, *args):
        """
        Raise a Not-implemented error, since this is not the way operations
        should be modified.
        """
        raise NotImplementedError(
            "Elements of the circuit cannot be substituted this way"
        )

    def __iter__(self):
        """Iterator protocol"""
        return iter(self._data)

    @property
    def num_sites(self):
        """Number of sites property"""
        return self._num_sites

    @property
    def num_qubits(self):
        """Number of sites property, with another name for simplicity of interfaces"""
        return self._num_sites

    @property
    def local_dim(self):
        """Local dimension property"""
        return self._local_dim

    @property
    def num_layers(self):
        """Number of layers in the circuit"""
        return len(self._data)

    @property
    def num_classical_sites(self):
        """Number of classical sites, i.e. bits"""
        num_cl = 0
        for _, item in self._cregisters.items():
            num_cl += len(item)
        return num_cl

    @property
    def cregisters(self):
        """Return the full list of classical registers"""
        return self._cregisters

    def _add_layer(self):
        """
        Add an empty layer to the circuit, so filled with wait operations
        """
        if len(self._occupation_map) == 0:
            self._occupation_map = np.zeros((self._num_sites, 1), dtype=bool)
        else:
            self._occupation_map = np.hstack(
                (self._occupation_map, np.zeros((self._num_sites, 1), dtype=bool))
            )
        self._data.append([])

    def inspect_cregister(self, register, idx=None):
        """
        Retrieve the decimal value of the binary number
        stored in the classical register `register` if idx=None,
        otherwise it returns the value of the bit at position idx

        Parameters
        ----------
        register : str
            Name of the classical register
        idx : int, optional
            Index of the bit you want to recover. If None, you will
            recover the decimal representation of the full register.
            Default to None.
        """
        if idx is None:
            binary_str = np.array2string(
                self._cregisters[register][:, 0][::-1].astype(int), separator=""
            )
            binary_str = binary_str.replace("[", "").replace("]", "")
            out = int(binary_str, 2)
        else:
            out = int(self._cregisters[register][idx, 0])

        return out

    def add_cregister(self, name, num_bits):
        """
        Add a classical register with name `name` and
        length `num_bits`

        Parameters
        ----------
        name : str
            Name of the classical register
        num_bits : int
            Number of bits the register can store
        """
        self._cregisters[name] = np.zeros((num_bits, 2), dtype=float)
        self._cl_occupation_map[name] = np.zeros(num_bits, dtype=int)

    def modify_cregister(self, value, name, idx):
        """
        Modify the value of a measured bit in the classical register

        Parameters
        ----------
        value : tuple of (int, float)
            Value of the measured qubit to be saved in the classical register
            and of the probability of that outcome
        name : str
            Name of the classical register
        idx : int
            Index of the bit where to save value
        """
        self._cregisters[name][idx, 0] = value[0]
        self._cregisters[name][idx, 1] = np.real(value[1])

    def add_site(self, new_register, position, reference_register="default"):
        """
        Add a site to the circuit, in a specific
        position. The position is the index of the link where
        the site will be added, so 0 is attached to the left,
        num_sites is attached to the right. An intermediate
        number attaches it in the middle.

        Parameters
        ----------
        new_register : str
            Name of the register that will track the index of
            the new site
        position : int
            Index of the link where the new site should be
            added in the register
        reference_register: str
            Register with respect to which the position is
            taken. Default to 'default'.
        """
        if not 0 <= position <= len(self._qregisters[reference_register]):
            raise ValueError(
                f"Site cannot be inserted in position {position}"
                + f" of register {reference_register} with"
                + f" num_sites={len(self._qregisters[reference_register])}"
            )

        # Recover true position on the MPS chain from the position in the register
        if position < len(self._qregisters[reference_register]):
            true_pos = self._qregisters[reference_register][position]
        else:
            true_pos = self._qregisters[reference_register][position - 1] + 1

        # Update all the values after that one, adding one to their
        # relative position
        for idxs in self._qregisters.values():
            mask = idxs >= true_pos
            idxs[mask] += 1

        # Insert the new site in the register
        if new_register in self._qregisters:
            idx = np.nonzero(self._qregisters[new_register] > true_pos)[0]
            if len(idx) == 0:
                self._qregisters[new_register] = np.append(
                    self._qregisters[new_register], true_pos
                )
            else:
                idx = idx[0]
                self._qregisters[new_register] = np.insert(
                    self._qregisters[new_register], idx[0], true_pos
                )
        # Add new register with the new index if it is next
        else:
            self._qregisters[new_register] = np.array([true_pos])

        # Insert a new row in the occupation map and update num_sites
        om_shape = self._occupation_map.shape[1]
        self._occupation_map = np.insert(
            self._occupation_map, true_pos, np.zeros(om_shape), axis=0
        )
        self._num_sites += 1

        # Put the information in the data
        self.add(QCAddSite(position), np.arange(self.num_sites), None)

    def remove_site(self, register, position):
        """
        Remove a site from the circuit. The site to be removed is
        identified by its register and the position in the register.

        Parameters
        ----------
        register : str
            Name of the register where the site to remove is
        position : int
            Index of the site to remove
        """
        if not 0 <= position < len(self._qregisters[register]):
            raise ValueError(
                f"Site cannot be removed in position {position}"
                + f" of register {register} with"
                + f" num_sites={len(self._qregisters[register])}"
            )

        # Put the information in the data
        self.add(QCRemoveSite(position), np.arange(self.num_sites), None)

        # Recover true position on the MPS chain from the position in the register
        true_pos = self._qregisters[register][position]

        # Update all the values after that one, subtracting one to their
        # relative position
        for idxs in self._qregisters.values():
            mask = idxs > true_pos
            idxs[mask] -= 1

        # Remove the site from the register
        self._qregisters[register] = np.delete(self._qregisters[register], position)

        # Remove row from the occupation map and update num_sites
        self._num_sites -= 1
        self._occupation_map = np.delete(self._occupation_map, true_pos, axis=0)

    def add_qregister(self, new_register, positions, reference_register="default"):
        """
        Add a new register to the circuit, in a specific
        positions. The positions is the index of the link where
        the site will be added, so 0 is attached to the left,
        num_sites is attached to the right. An intermediate
        number attaches it in the middle.
        This link number is taken with respect to the
        reference register.

        This function is a simple loop over add_site, but we expect
        it to be a very useful function in runtime

        Parameters
        ----------
        new_register : str
            Name of the register that will track the index of
            the new site
        positions : array-like of ints
            Index of the link where the new site should be
            added in the register
        reference_register: str
            Register with respect to which the position is
            taken. Default to 'default'.
        """
        for position in positions:
            self.add_site(new_register, position, reference_register)

    def remove_qregister(self, register):
        """
        Remove a register from the circuit.

        Parameters
        ----------
        register : str
            Name of the register to be removed
        """
        num_reps = len(self._qregisters[register])
        for _ in range(num_reps):
            self.remove_site(register, 0)

        # Remove the key from the dictionary
        del self._qregisters[register]

    # pylint: disable-next=too-many-branches
    def add(self, operation, sites, register="default"):
        """
        Add an operation to the circuit

        Parameters
        ----------
        operation : :class:`QCOperation`
            Operation to be added to the
        sites : integer of array-like of integers
            Qubits to which the operation is applied
        register : str or array of str
            Register to which the operations are applied.
            Default are both to default.
        """
        if not issubclass(type(operation), QCOperation):
            raise TypeError(
                f"operation must be of type QCOperation, not {type(operation)}"
            )
        if isinstance(sites, int):  # Trasnsfor integer in array-like list
            sites = [sites]
            self._num_one_site_operations += 1
        else:
            self._num_two_site_operations += 1
        if isinstance(register, str):
            register = [register] * len(sites)

        if register is not None:
            for idx, site in enumerate(sites):
                sites[idx] = self._qregisters[register[idx]][site]

        sites = list(sites)
        last_occupied = -1
        for site in sites:
            temp = np.where(self._occupation_map[site, :])[0]
            temp = [-1] if len(temp) == 0 else temp
            last_occupied = max(last_occupied, temp[-1])

        # Check for classical dependencies
        if operation.is_conditioned:
            if operation.c_if.idx is None:
                for occ in self._cl_occupation_map[operation.c_if.creg]:
                    last_occupied = max(last_occupied, occ)
                self._cl_occupation_map[operation.c_if.creg] = last_occupied + 1
            else:
                last_occupied = max(
                    last_occupied,
                    self._cl_occupation_map[operation.c_if.creg][operation.c_if.idx],
                )
                self._cl_occupation_map[operation.c_if.creg][operation.c_if.idx] = (
                    last_occupied + 1
                )
        # Check for measures
        if hasattr(operation, "cregister"):
            occ = self._cl_occupation_map[operation.cregister][operation.cl_idx]
            last_occupied = max(last_occupied, occ)
            self._cl_occupation_map[operation.cregister][operation.cl_idx] = (
                last_occupied + 1
            )

        # Check if the last occupied is on the last layer,
        # and add a layer if True
        if last_occupied == self._occupation_map.shape[1] - 1:
            self._add_layer()

        # Add the operation on data and occupation map
        instruction = [operation, sites]
        self._data[last_occupied + 1].append(instruction)
        self._occupation_map[sites, last_occupied + 1] = True
        # print(self._occupation_map)

    # pylint: disable-next=too-many-arguments
    def insert(self, operation, sites, layer, register="default", inlayer=True):
        """
        Insert an operation after a specific layer on the sites
        If inlayer is True simply add it into the layer without
        checking if an operation is already there, otherwise check
        for the operation and insert a new layer

        Parameters
        ----------
        operation : :class:`QCOperation`
            Operation to be added to the
        sites : int of array-like of integers
            Qubits to which the operation is applied
        layer: int
            integer identifying the layer after which we want to
            insert the operation
        register : str or array of str
            Register to which the operations are applied.
            Default are both to default.
        inlayer: bool, optional
            If False, and the position in the layer is already occupied,
            add a new layer. Otherwise add it to that layer nevertheless.
            Default to True.
        """
        if not issubclass(type(operation), QCOperation):
            raise TypeError(
                f"operation must be of type QCOperation, not {type(operation)}"
            )
        if np.isscalar(sites):  # Trasnsfor integer in array-like list
            sites = [sites]
        if isinstance(register, str):
            register = [register] * len(sites)

        for idx, site in enumerate(sites):
            sites[idx] = self._qregisters[register[idx]][site]

        instruction = [operation, sites]
        # Check if the positions sites in layer is occupied
        is_occupied = self._occupation_map[sites, layer]
        if not np.isscalar(is_occupied):
            is_occupied = any(is_occupied)

        if (not inlayer) and is_occupied:
            # Insert new layer in occupation map and data
            self._occupation_map = np.insert(
                self._occupation_map, layer + 1, False, axis=1
            )
            self._data.insert(layer + 1, [])
            self._occupation_map[sites, layer + 1] = True
            layer += 1
        elif not inlayer and not is_occupied:
            self._occupation_map[sites, layer] = True
        # Insert instruction
        self._data[layer].append(instruction)

    @classmethod
    def from_qiskit(cls, qc):
        """
        Initialize the Qcircuit class from a qiskit quantum circuit

        Parameters
        ----------
        qc : :class:`QuantumCircuit`
            qiskit QuantumCircuit that will initialize the Qcircuit
        """
        if not isinstance(qc, QuantumCircuit):
            raise TypeError(f"qc must be of type QuantumCircuit, not {type(qc)}")

        # Reinitialize values
        qcirc = cls(qc.num_qubits)

        for instance in qc.data:
            gate = instance[0]
            if gate.name == "barrier":
                pass
            elif gate.name == "measure":
                qcirc.measure_projective(
                    [qub.index for qub in instance[1]],
                    [cl_idx.index for cl_idx in instance[2]],
                )
            else:
                operation = QCOperation(gate.name, gate.to_matrix)
                qcirc.add(operation, sites=[qub.index for qub in instance[1]])

        return qcirc

    def fill_qcoperators(self, operators):
        """
        Fill the QCoperators dictionary with all the operators
        in the Qcircuit

        Parameters
        ----------
        operators: :class:`QCOperators`
            operators class to fill
        """
        if not isinstance(operators, QCOperators):
            raise TypeError(f"qc must be of type QuantumCircuit, not {type(operators)}")

        for layer in self._data:
            for instruction in layer:
                operation = instruction[0]
                sites = instruction[1]
                name = operation.name
                matrix = operation.operator
                search = True
                while search:
                    # If the operator has already been written, even if parametric,
                    # don't add it again. Instead, if it has sufficiently different
                    # parameters, add it with a new name
                    if name in operators:
                        ii = 0
                        if np.allclose(matrix, operators[name]):
                            search = False
                        else:
                            name = operation.name + str(ii)
                            ii += 1
                    else:
                        search = False
                        operators[name] = matrix.reshape([self.local_dim] * len(sites))
                        self._namelist.append(name)

    def write(self, dest, id_mapping):
        """
        Write on file (originally written for the Fortran simulator), with the correct formatting

        Parameters
        ----------
        dest: str or filehandle
            If str, write on that file. Otherwise, write using that filehandle
        id_mapping: dict
            Mapping from the operator name to the operator id
        """
        if isinstance(dest, str):
            fh = open(dest, "w+")
        elif hasattr(dest, "write"):
            fh = dest
        else:
            raise TypeError("`dest` for `write` neither str nor writeable.")

        fh.write(str(self.num_sites) + "\n")
        fh.write(str(self.num_classical_sites) + "\n")
        fh.write(str(len(self) + len(self._data)) + "\n")

        ii = 0
        for layer in self._data:
            for instruction in layer:
                operation = instruction[0]
                sites = instruction[1]
                fh.write(
                    operation.name, id_mapping(self._namelist[ii]), len(sites), 0, "\n"
                )
                fh.write("\t")
                for site in sites:
                    fh.write(str(site + 1) + " ")  # Qubit numbers
                for clbit in range(0):
                    fh.write(str(clbit + 1) + " ")  # Classical bit numbers
                fh.write("\n")
                ii += 1
            fh.write("barrier", "1", "1", "1", "\n", "\n")

    # pylint: disable-next=too-many-branches
    def to_matrix(self, sparse=False, max_qubit_equivalent=20, qiskit_order=True):
        """
        Return the matrix representation of the circuit.
        It is a unitary matrix if no noise/non-unitary matrix
        has been applied. Notice that this computation scales as
        :math:`d^{2n}`, where :math:`d` is the local dimension of the
        single degree of freedom and :math:`n` the number of sites.
        Furthermore, the function is not optimized and might take long.

        Parameters
        ----------
        sparse : bool, optional
            If True, return a csr sparse matrix instead of a numpy array.
            Default to False.
        max_qubit_equivalent: int, optional
            Maximum number of qubit sites the MPS can have and still be
            transformed into a matrix.
            If the number of sites is greater, it will throw an exception.
            Default to 20.
        qiskit_order : bool, optional
            If True, return the matrix using qiskit ordering. Default to
            True.

        Returns
        -------
        matrix : np.ndarray or sp.csr_matrix
            Numpy array or sparse matrix representing the system unitary
        """
        if self.local_dim**self.num_sites > 2**max_qubit_equivalent:
            raise RuntimeError(
                "Maximum number of sites for the matrix is "
                + f"fixed to the equivalent of {max_qubit_equivalent} qubit sites"
            )

        # Define the identity call based on the sparse flag
        if sparse:
            iden = sp.identity
            kron = sp.kron
            swap = sp.csr_matrix(QCSwap().operator)
        else:
            iden = np.identity
            kron = np.kron
            swap = QCSwap().operator

        matrix = iden(self.local_dim**self.num_sites)

        for layer in self._data:
            for instruction in layer:
                op_mat = instruction[0].operator
                if sparse:
                    op_mat = sp.csr_matrix(op_mat)
                sites = instruction[1]
                diff = max(sites) - min(sites)

                if len(sites) == 2:
                    # Create matrix if non adjacent
                    if diff > 1:
                        circ = Qcircuit(diff + 1, self.local_dim)
                        for ii in range(diff - 1):
                            circ.add(QCSwap(), [ii, ii + 1])
                        if sites[0] < sites[1]:
                            circ.add(instruction[0], [diff - 1, diff])
                        else:
                            circ.add(instruction[0], [diff, diff - 1])
                        for ii in range(diff - 1, 0, -1):
                            circ.add(QCSwap(), [ii, ii - 1])
                        op_mat = circ.to_matrix(sparse=sparse)

                    # Revert the direction if target is above control
                    elif sites[0] < sites[1]:
                        op_mat = swap @ op_mat @ swap

                # Reverse all the indexing if the qiskit order is required
                if qiskit_order:
                    sites = self.num_sites - 1 - np.array(sites)
                    if op_mat.shape == (4, 4):
                        op_mat = swap @ op_mat @ swap

                if min(sites) > 0:
                    op_mat = kron(iden(2 ** (min(sites))), op_mat)
                if max(sites) < self.num_sites - 1:
                    op_mat = kron(op_mat, iden(2 ** (self.num_sites - max(sites) - 1)))

                matrix = op_mat @ matrix

        return matrix

    # pylint: disable-next=too-many-locals
    def str_repr(self):
        """
        Get the string representation of the circuit
        """

        # Initialization
        str_repr = []
        biggest_idx_len = len(f"q_{self.num_sites-1} ")
        for idx in range(self.num_sites):
            str_idx = f"q_{idx} "
            str_idx += " " * (biggest_idx_len - len(str_idx))
            void = " " * biggest_idx_len

            str_repr.append([void, str_idx, void])

        for layer in self._data:
            used_sites = []
            max_len = 0
            for instruction in layer:
                op_str = instruction[0].string_rep.split("\n")
                sites = instruction[1]
                used_sites += list(sites)

                for idx, site in enumerate(sites):
                    current_str = op_str[idx * 3 : (idx + 1) * 3]
                    if len(current_str) == 0:
                        current_str = op_str[:3]

                    for idx_str, str1 in enumerate(current_str):
                        str_repr[site][idx_str] += str1
                        max_len = max(max_len, len(str1))

            empty_idxs = np.setdiff1d(np.arange(self.num_sites), np.array(used_sites))
            for site in empty_idxs:
                str_repr[site][0] += " " * max_len
                str_repr[site][1] += "─" * max_len
                str_repr[site][2] += " " * max_len
        circ_rep = ""
        for site in str_repr:
            circ_rep += "\n".join(site)

        return circ_rep

    ################################################################
    ## Dummy methods for the most used operations to simplify the ##
    ## life of the programmer. You could still use the add method ##
    ## since this is simply a dummy rewriting.                    ##
    ################################################################

    def x(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a x gate in the position pos in the qregister qreg

        Parameters
        ----------
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCXpauli(c_if), pos, qreg)

    def y(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a y gate in the position pos in the qregister qreg

        Parameters
        ----------
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCYpauli(c_if), pos, qreg)

    def z(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a z gate in the position pos in the qregister qreg

        Parameters
        ----------
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCZpauli(c_if), pos, qreg)

    def h(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a hadamard gate in the position pos in the qregister qreg

        Parameters
        ----------
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCHadamard(c_if), pos, qreg)

    def s(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a S gate in the position pos in the qregister qreg

        Parameters
        ----------
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCSgate(c_if), pos, qreg)

    def sadg(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply the adjoint of an S gate in the position pos
        in the qregister qreg

        Parameters
        ----------
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCSadggate(c_if), pos, qreg)

    def t(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a T gate in the position pos in the qregister qreg

        Parameters
        ----------
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCTgate(c_if), pos, qreg)

    def tadg(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply the adjoint of an T gate in the position pos
        in the qregister qreg

        Parameters
        ----------
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCTadggate(c_if), pos, qreg)

    def p(self, theta, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a Phase gate in the position pos in the qregister qreg
        with a phase theta

        Parameters
        ----------
        theta : float
            Rotation angle for the phase gate
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCPgate(theta, c_if), pos, qreg)

    def cx(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a controlled not gate in the position pos
        in the qregisters qreg. pos should be an array, qreg should
        be an array if the qubits are in different qregisters

        Parameters
        ----------
        pos : list of two ints
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : list of str, str or None, optional
            Quantum registers where you want to apply the operation.
            If only a str is provided, the program assumes you want to
            apply the gate inside the same quantum register. If a list of
            two strings is provided, the gate is applied between different
            quantum registers.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'.
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCCnot(c_if), pos, qreg)

    def cz(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a controlled z gate in the position pos
        in the qregisters qreg. pos should be an array, qreg should
        be an array if the qubits are in different qregisters

        Parameters
        ----------
        pos : list of two ints
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : list of str, str or None, optional
            Quantum registers where you want to apply the operation.
            If only a str is provided, the program assumes you want to
            apply the gate inside the same quantum register. If a list of
            two strings is provided, the gate is applied between different
            quantum registers.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'.
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCCz(c_if), pos, qreg)

    def cp(self, theta, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a controlled phase gate in the position pos
        in the qregisters qreg. pos should be an array, qreg should
        be an array if the qubits are in different qregisters

        Parameters
        ----------
        theta : float
            Rotation angle for controlled phase gate
        pos : list of two ints
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : list of str, str or None, optional
            Quantum registers where you want to apply the operation.
            If only a str is provided, the program assumes you want to
            apply the gate inside the same quantum register. If a list of
            two strings is provided, the gate is applied between different
            quantum registers.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'.
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCCp(theta, c_if), pos, qreg)

    def swap(self, pos, qreg="default", c_if=ClassicalCondition()):
        """
        Apply a swap gate in the position pos
        in the qregisters qreg. pos should be an array, qreg should
        be an array if the qubits are in different qregisters

        Parameters
        ----------
        pos : list of two ints
            Index, relative to the quantum register, where you want to
            apply the operation
        qreg : list of str, str or None, optional
            Quantum registers where you want to apply the operation.
            If only a str is provided, the program assumes you want to
            apply the gate inside the same quantum register. If a list of
            two strings is provided, the gate is applied between different
            quantum registers.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'.
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCSwap(c_if), pos, qreg)

    # pylint: disable-next=too-many-arguments
    def gate(self, gate, pos, name="op", qreg="default", c_if=ClassicalCondition()):
        """
        Apply a matrix to the circuit. The dimensions of
        the matrix should be consistent with the number of
        qubits to which it should be applied

        Parameters
        ----------
        gate : np.ndarray
            gate matrix
        pos : int or array-like of ints
            Indexes of the sites to which it should be applied
        name : str, optional
            Unique identifier of the gate. Default to 'op'
        qreg : str or array-like of str, optional
            qregister to which the gate should be applied
        """
        if isinstance(pos, int):
            pos = [pos]

        if gate.shape[1] != self.local_dim ** len(pos):
            raise ValueError(
                f"Gate shape {gate.shape[1]} is incompatible "
                + "with the local dimension and the number of sites to which "
                + f"it is applied. Should be {self.local_dim**len(pos)}"
            )
        op = QCOperation(name, lambda: gate, conditioned=c_if)
        self.add(op, pos, qreg)

    # pylint: disable-next=too-many-arguments
    def measure_projective(
        self,
        pos,
        cl_idx,
        qreg="default",
        creg="default",
        selected_output=None,
        c_if=ClassicalCondition(),
    ):
        """
        Apply a projective measurement in the position pos
        in the qregister qreg. You can choose the output of
        the measure. The result is stored in the classical
        qregister creg at index cl_idx

        Parameters
        ----------
        pos : int
            Index, relative to the quantum register, where you want to
            apply the operation
        cl_idx : int
            Index of the classical register where to store the result
            of the measurement
        qreg : str or None, optional
            Quantum register where you want to apply the operation.
            If None, the "raw index" is considered, i.e. without taking
            into account the quantum register reindexing.
            Default to 'default'
        cregister : str, optional
            Classical register where the result of the measurement
            is stored. Default to 'default'
        selected_output : int, optional
            Output selected a priori. Default to None.
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation
            on the simulation.
        """
        self.add(QCMeasureProjective(cl_idx, selected_output, creg, c_if), pos, qreg)

    def renormalize(self, c_if=ClassicalCondition()):
        """
        Renormalize the state

        Parameters
        ----------
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation on the
            simulation
        """
        self.insert(QCRenormalize(c_if), 0, -1)

    # pylint: disable-next=too-many-arguments
    def measure_observables(
        self, name, observables, operators, qreg=None, c_if=ClassicalCondition()
    ):
        """
        Perform a measurement step of all the observables defined in
        the observable parameter.

        Parameters
        ----------
        name : str
            Name of the observables step
        observables: :class:`TNObservables`
            Observables to be measured at this step
        operators: :class:`QCOperators`
            Class containing the operators present in the Observables
        qreg : str or None, optional
            Quantum register where the observables should be measured.
            If None, they are measured on the FULL circuit, regardless
            of the registers
        c_if : :py:class:`ClassicalCondition`, optional
            Classical condition to be checked for applying the operation on the
            simulation
        """
        obs = QCObservableStep(name, observables, operators, c_if)
        obs.num_sites = self.num_sites

        # We want to apply the measurement to the whole circuit regardless
        # of the register
        if qreg is None:
            obs.snap_current_qregisters(self._qregisters)
            self.add(obs, np.arange(self.num_sites - 1), register=qreg)
        # We want to apply the measurements to a single part of the circuit
        else:
            qreg_val = deepcopy(self._qregisters[qreg])
            obs.snap_current_qregisters({qreg: qreg_val})
            # Adjust the index of the observables based on specific sites
            obs.adjust_obs_indexing()
            # Add occupying the whole interested quantum register
            self.add(obs, np.arange(self.num_sites - 1), register=None)



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
In order to simulate the quantum circuit we first need to preprocess it, to put it in a way
suitable for the MPS simulation.

Preprocessing
~~~~~~~~~~~~~

The tensor network ansatz that is used in the simulator is called Matrix Product States (MPS).
In this ansatz each degree of freedom (qubit, mode, particle) is treated as a rank-3 tensor.
We can only apply operators which are **local** or that applies to **nearest neighbors**.
This means that, when a circuit is passed to the preprocessing it is translated in an equivalent
circuit that follows these two properties. While for quantum optics these properties are already
ensured, for qubits quantum circuit we have to perform some transformations. We use qiskit for them,
and in particular we:

- Map the circuit using a pre-defined set of gates, called basis_gates.
  These gates can be passed to the function, to satisfy
  a particular constraint of a physical machine we want to simulate
- Map the circuit into a linear circuit. By an optimized application of *swap*
  gates we map non-local two-qubit gates into a series
  of local two-qubit gates.

These operations are automatically performed by the higher level function :py:func:`run_simulation`.
However, it is important to take into account that these operations require time, which is
proportional to the size of the circuit. For this reason, if the user has to perform multiple
simulation of the same circuit varing other parameres, e.g. the bond dimension, it is suggested to
perform the preprocessing only once, and then disable the mapping and the linearization of the
function.

.. warning::
    The preprocessing and linearization using qiskit modifies the order of the qubits.
    This is a behavior we don't want to experience, and such inside the preprocessing
    procedure the qubits are again ordered in the initial layout.

"""

from qiskit import QuantumCircuit, transpile, transpiler

from .tensor_compiler import tensor_compiler
from .utils.qk_utils import get_index, qk_transpilation_params

__all__ = ["preprocess"]


def _bubble_sort(nums):
    """Given an integer array sort it by only swapping nearest neighbors elements,
        and return the necessary swaps to bring the array into sorted shape

    Parameters
    ----------
    nums : list of int
        Array of integer to be sorted in ascendant order

    Returns
    -------
    swaps : list of tuples of int
        Indexes of the qubits that has to be swapped
    """
    # We set swapped to True so the loop looks runs at least once
    swapped = True
    swaps = []
    while swapped:
        swapped = False
        for ii in range(len(nums) - 1):
            if nums[ii] > nums[ii + 1]:
                # Swap the elements
                nums[ii], nums[ii + 1] = nums[ii + 1], nums[ii]
                # Set the flag to True so we'll loop again
                swapped = True
                swaps.append((ii, ii + 1))
    return swaps


def _reorder_qk(linearized_qc):
    """Qiskit transpiler does not control the final layout of the qubits, i.e.
        we don't know if they are ordered correctly. It takes care of this problem
        by rearrenging the measurements on the classical register. However, it is
        a problem for the mps simulator. We so reorder the qubit register by applying
        swap gates.

    Parameters
    ----------
    linearized_qc : :py:class:`qiskit.QuantumCircuit`
        Circuit linearized trhough the qiskit transpiler

    Returns
    -------
    linearized_qc : :py:class:`qiskit.QuantumCircuit`
        Circuit with the correct ending qubits
    """
    num_qub = linearized_qc.num_qubits
    # Read the final layout by the rearrengement of the measurements
    final_perm = [0] * num_qub
    for instruction in linearized_qc.data[linearized_qc.size() - num_qub + 1 :]:
        if instruction.operation.name == "measure":
            final_perm[get_index(linearized_qc, instruction.qubits[0])] = get_index(
                linearized_qc, instruction.clbits[0]
            )
    # Get the combination of swaps necessary trhough the bubble_sort algorithm,
    # which is O(num_qub^2)
    swaps = _bubble_sort(final_perm)
    # Perform the swapping operations
    linearized_qc.remove_final_measurements()
    for ii, jj in swaps:
        linearized_qc.swap(ii, jj)

    return linearized_qc


def _preprocess_qk(qc, qk_params=qk_transpilation_params()):
    """Transpile the circuit to adapt it to the linear structure of the MPS, with the constraint
    of having only the gates basis_gates

     Parameters
    ----------
    qc: QuantumCircuit
         qiskit quantum circuit
    linearize: bool, optional
        If True use qiskit transpiler to linearize the circuit. Default to True.
    basis_gate: list, optional
        If not empty decompose using qiskit transpiler into basis gate set
    optimization: intger, optional
        Level of optimization in qiskit transpiler. Default to 3.

    Returns
    -------
    linear_qc: QuantumCircuit
        Linearized quantum circuit
    """
    # Empty circuit case
    if len(qc.data) == 0:
        return qc
    basis_gates = [] if qk_params.basis_gates is None else qk_params.basis_gates
    n_qub = qc.num_qubits
    linear_map = transpiler.CouplingMap.from_line(n_qub)
    # Assure that final measurements are present, by first eliminating and adding them
    qc.remove_final_measurements()
    qc.measure_all()
    # Transpile the circuit
    if len(basis_gates) > 0 and qk_params.linearize:
        linear_qc = transpile(
            qc,
            coupling_map=linear_map,
            optimization_level=qk_params.optimization,
            basis_gates=basis_gates,
            initial_layout=list(range(n_qub)),
        )
    elif len(basis_gates) == 0 and qk_params.linearize:
        linear_qc = transpile(
            qc,
            coupling_map=linear_map,
            optimization_level=qk_params.optimization,
            initial_layout=list(range(n_qub)),
        )
    elif len(basis_gates) > 0:
        linear_qc = transpile(
            qc,
            optimization_level=qk_params.optimization,
            basis_gates=basis_gates,
            initial_layout=list(range(n_qub)),
        )
    else:
        linear_qc = qc

    # Reorder the circuit
    linear_qc = _reorder_qk(linearized_qc=linear_qc)

    # Use the tensor compiler if requested
    if qk_params.tensor_compiler:
        linear_qc = tensor_compiler(linear_qc)

    return linear_qc


def preprocess(circ, **kwargs):
    """
    Interface for the preprocessing of qiskit circuits

    Parameters
    ----------
    circ: qiskit.circuit.quantumcircuit.QuantumCircuit
        Quantum circuit instance
    **kwargs:
        linearize: bool, optional
            If True use qiskit transpiler to linearize the circuit. Default to True.
        basis_gate: list, optional
            If not empty decompose using qiskit transpiler into basis gate set
        optimization: integer, optional
            Level of optimization in qiskit transpiler. Default to 3.

    Returns
    -------
    preprocessed_circ: qiskit.circuit.quantumcircuit.QuantumCircuit
        Preprocessed quantum circuit instance
    """

    if isinstance(circ, QuantumCircuit):
        preprocessed_circ = _preprocess_qk(circ, **kwargs)

    else:
        raise TypeError(
            "Only qiskit quantum circuit are implemented. Your circuit is of type:"
            + str(type(circ))
        )

    return preprocessed_circ



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os
import os.path
import sys
import unittest
from shutil import rmtree

import numpy as np

# Try to import cupy
try:
    import cupy as cp
    from cupy_backends.cuda.api.runtime import CUDARuntimeError

    try:
        _ = cp.cuda.Device()
        GPU_AVAILABLE = True
    except CUDARuntimeError:
        GPU_AVAILABLE = False
except ImportError:
    cp = None
    GPU_AVAILABLE = False

from qiskit import ClassicalRegister, QuantumCircuit, transpile, transpiler
from qiskit.circuit.library import QuantumVolume
from qtealeaves.observables import TNObservables, TNObsProjective, TNState2File

from qmatchatea import QCBackend, run_simulation
from qmatchatea.circuit import Qcircuit
from qmatchatea.utils import QCIO, QCConvergenceParameters, fidelity
from qmatchatea.utils.mpi_utils import to_layered_circ
from qmatchatea.utils.qk_utils import (
    GHZ_qiskit,
    W_qiskit,
    qiskit_get_statevect,
    qk_transpilation_params,
)

sys.argv = [""]
avail_gates = [
    "x",
    "y",
    "z",
    "h",
    "id",
    "s",
    "sdg",
    "sx",
    "sxdg",
    "t",
    "tdg",
    "swap",
    "dcx",
    "ecr",
    "iswap",
    "ch",
    "cx",
    "cy",
    "cz",
    "p",
    "r",
    "rx",
    "ry",
    "rz",
    "u",
    "u1",
    "u2",
    "rxx",
    "ryy",
    "rzx",
    "rzz",
    "cp",
    "cry",
    "crz",
    "cu",
    "cu1",
]


def qvol_circ(num_qub=3):
    qc = QuantumVolume(num_qub)
    qc = qc.decompose()
    lin_map = transpiler.CouplingMap.from_line(num_qub)
    qc = transpile(qc, coupling_map=lin_map, basis_gates=avail_gates)
    qc = to_layered_circ(qc)
    # qc.add_register(creg)

    return qc


class TestBasicCircuits(unittest.TestCase):
    def setUp(self):
        if not os.path.isdir("TMP_TEST"):
            os.makedirs("TMP_TEST")
        self.qcio = QCIO("TMP_TEST/data/in", "TMP_TEST/data/out")
        self.backends = self.get_python_backends()
        self.obs = TNObservables()
        self.obs += TNObsProjective(16)
        self.obs += TNState2File("TMP_TEST/data/out/statevector.txt", "F")
        np.random.seed(123)

    @staticmethod
    def get_python_backends():
        """
        Generate all possible backends for python.
        """
        backends = (
            {
                "precision": "C",
                "device": "cpu",
                "ansatz": "MPS",
            },
            {
                "precision": "Z",
                "device": "cpu",
                "ansatz": "MPS",
            },
            {
                "precision": "C",
                "device": "cpu",
                "ansatz": "TTN",
            },
            {
                "precision": "Z",
                "device": "cpu",
                "ansatz": "TTN",
            },
            {
                "precision": "C",
                "device": "gpu",
                "ansatz": "MPS",
            },
            {
                "precision": "Z",
                "device": "gpu",
                "ansatz": "MPS",
            },
        )

        if GPU_AVAILABLE:
            pass
        else:
            backends = (backends[0], backends[1], backends[2], backends[3])

        backends = [QCBackend(**elem) for elem in backends]

        return backends

    def tearDown(self):
        if os.path.isdir("TMP_TEST"):
            rmtree("TMP_TEST")
        return

    def test_GHZ(self):
        """
        Check that the GHZ circuit is reproduced correctly
        """
        for backend in self.backends:
            num_qubits = 8
            qc = QuantumCircuit(num_qubits)
            GHZ_qiskit(qc)
            true_state = qiskit_get_statevect(qc)
            res = run_simulation(
                qc, io_info=self.qcio, backend=backend, observables=self.obs
            )
            statevect = res.statevector
            fid = fidelity(statevect, true_state)
            self.assertAlmostEqual(
                fid, 1, places=12, msg="GHZ state not described correctly"
            )

        return

    def test_GHZ_qcircuit(self):
        """
        Check that the GHZ circuit is reproduced correctly
        """
        num_qubits = 8
        true_state = np.zeros(2**num_qubits)
        true_state[[0, -1]] = 1 / np.sqrt(2)
        for backend in self.backends:
            qc = Qcircuit(num_qubits)
            qc.h(0)
            for ii in range(0, num_qubits - 1):
                qc.cx([ii, ii + 1])
            res = run_simulation(
                qc, io_info=self.qcio, backend=backend, observables=self.obs
            )
            statevect = res.statevector
            fid = fidelity(statevect, true_state)
            self.assertAlmostEqual(
                fid, 1, places=12, msg="GHZ state not described correctly with Qcircuit"
            )

        return

    def test_W(self):
        """
        Check that the W circuit is reproduced correctly
        """
        for backend in self.backends:
            num_qubits = 8
            qc = QuantumCircuit(num_qubits)
            W_qiskit(qc)
            true_state = qiskit_get_statevect(qc)
            res = run_simulation(
                qc, io_info=self.qcio, backend=backend, observables=self.obs
            )
            statevect = res.statevector
            fid = fidelity(statevect, true_state)
            self.assertAlmostEqual(
                fid, 1, places=12, msg="W state not described correctly"
            )

        return

    def test_QVOLUME(self):
        """
        Check that the QVOLUME circuit is reproduced correctly
        """
        for backend in self.backends:
            approach = backend.identifier

            num_qubits = 8
            transpilation_params = qk_transpilation_params(linearize=True)
            convergence_params = QCConvergenceParameters(max_bond_dimension=50)
            qc = qvol_circ(num_qubits)
            true_state = qiskit_get_statevect(qc)
            res = run_simulation(
                qc,
                convergence_parameters=convergence_params,
                io_info=self.qcio,
                transpilation_parameters=transpilation_params,
                backend=backend,
                observables=self.obs,
            )
            statevect = res.statevector
            fid = fidelity(statevect, true_state)
            tol_places = 12 if (backend.precision == "Z") else 5
            self.assertAlmostEqual(
                fid,
                1,
                places=tol_places,
                msg=f"QVOLUME state not described correctly with {approach}.",
            )

        return



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
Compiler for a qiskit quantum circuit in an optimized form for a tensor network simulator.
Since the gates are contracted with the MPS the naive structure below consists of 4
contractions with the MPS and 3 SVDs. These operations are done with
:math:`\chi\times\chi\times2` tensors, and such are computationally demanding. Instead,
we first contract all these gates together, performing at most matrix-matrix multiplications
with :math:`4\times4` matrices.

.. code-block::

    q_0 --G--G--G--G--
          |     |  |
    q_1 --G-----G--G--

We indicated with G a generic gate.

Functions and classes
---------------------

"""

import numpy as np
from qiskit.circuit.library import UnitaryGate
from qiskit.converters import circuit_to_dag, dag_to_circuit

from .utils.qk_utils import get_index

__all__ = ["tensor_compiler"]


def are_contractable(this, other):
    """
    Check if this DagNode ant the other DagNode
    are contractable, i.e. if they act on the
    same qubits

    Parameters
    ----------
    this : DagNode
        First node
    other : DagNode
        second node

    Returns
    -------
    int
        0: they are not contractible
        1: they are contractible and acts on the
           same number of qubits
        2: they are contractible and acts on different
           number of qubits
    """
    if (
        (this.op.condition is not None)
        or len(this.cargs)
        or (this.op.name == "swap") != 0
    ):
        return 0

    # Same number of qubits
    if len(this.qargs) == len(other.qargs):
        if all(qq in other.qargs for qq in this.qargs):
            return 1
    # The second is a one-qubit gate
    elif len(other.qargs) == 1:
        one_qub = other.qargs[0]
        if one_qub in other.qargs:
            return 2
    # The first is a one-qubit gate
    elif len(this.qargs) == 1:
        one_qub = this.qargs[0]
        if one_qub in other.qargs:
            return 3

    return 0


def invert_qiskit_matrix(matrix):
    """
    Invert a matrix between qiskit
    and non-qiskit convention of the
    numbering.
    (qiskit orders qubit opposite to our convention,
    swap inside rows and columns)

    Parameters
    ----------
    matrix : np.ndarray, shape (4,4)
        Matrix in qiskit/non-qiskit format

    Return
    ------
    np.ndarray, shape(4, 4)
        Matrix in non-qiskit/qiskit format
    """
    # Reshape into tensor, perform correct permutation
    matrix = matrix.reshape(2, 2, 2, 2).transpose([1, 0, 3, 2])
    # Return matrix
    return matrix.reshape(4, 4)


def contract(qc, this, other):
    """
    Contract this DagNode and the other DagNode
    in the quantum circuit qc

    Parameters
    ----------
    qc : QuantumCircuit
        Quantum circuit where the DagNodes are defined
    this : DagNode
        First node
    other : DagNode
        second node

    Returns
    -------
    int
        The contraction flag
    DagNode
        The contraction of this and other
    """

    # Check if the two instructions are contractible
    flag = are_contractable(this, other)

    # No contraction allowed
    if flag == 0:
        return flag, this.op

    # Contraction allowed
    this_tens = this.op.to_matrix()
    other_tens = other.op.to_matrix()
    this_qargs = [get_index(qc, qub) for qub in this.qargs]
    other_qargs = [get_index(qc, qub) for qub in other.qargs]

    # This and other have same number of qubits
    if flag == 1:
        # Two one-qubit gates
        if len(this.qargs) == 1:
            final_tens = other_tens @ this_tens

        # Two two-qubits gates
        else:
            if this_qargs[0] != other_qargs[0]:
                other_tens = invert_qiskit_matrix(other_tens)

            final_tens = other_tens @ this_tens

    # This and other acts on a different number of qubits
    elif flag == 2:
        # True if the first qubit gate is on the first qubit
        gate_on_first_qub = other_qargs[0] == min(this_qargs)
        # True if the control qubit is the first
        up_down = this_qargs[0] < this_qargs[1]
        # Promote single qubit to single qubit tensor identity
        if gate_on_first_qub ^ up_down:  # Logical xor
            other_tens = np.kron(other_tens, np.eye(2, 2))
        else:
            other_tens = np.kron(np.eye(2, 2), other_tens)

        final_tens = other_tens @ this_tens

    elif flag == 3:
        # True if the first qubit gate is on the first qubit
        gate_on_first_qub = this_qargs[0] == min(other_qargs)
        # True if the control qubit is the first
        up_down = other_qargs[0] < other_qargs[1]
        # Promote single qubit to single qubit tensor identity
        if gate_on_first_qub ^ up_down:  # Logical xor
            this_tens = np.kron(this_tens, np.eye(2, 2))
        else:
            this_tens = np.kron(np.eye(2, 2), this_tens)

        final_tens = other_tens @ this_tens
    else:
        raise ValueError(f"Value for flag={flag}, but not implemented.")

    return flag, UnitaryGate(final_tens)


def contract_1q_runs(qc, dag):
    """
    Contract subsequent one qubit
    gates in the quantum circuit qc

    Parameters
    ----------
    qc : QuantumCircuit
        quantum circuit of interest
    dag : DagCircuit
        Dag representation of qc

    Returns
    -------
    DagCircuit
        new dag circuit with the single qubits
        contracted
    """
    for run in dag.collect_1q_runs():
        new_node = run[0]

        for node in run[1:]:
            flag, instr = contract(qc, new_node, node)

            if flag == 1:
                dag.remove_op_node(node)
                new_node = dag.substitute_node(new_node, instr, True)
    return dag


def contract_2q_runs(qc, dag):
    """
    Contract subsequent two qubit
    gates in the quantum circuit qc

    Parameters
    ----------
    qc : QuantumCircuit
        quantum circuit of interest
    dag : DagCircuit
        Dag representation of qc

    Returns
    -------
    DagCircuit
        new dag circuit with the two-qubits
        gates contracted
    """
    skip_next = False
    for run in dag.collect_2q_runs():
        new_node = run[0]
        for idx, node in enumerate(run[1:]):
            if skip_next:
                skip_next = False
                continue
            flag, instr = contract(qc, new_node, node)

            if flag == 0:
                new_node = run[idx + 1]
                continue

            if flag == 1:
                dag.remove_op_node(node)
                new_node = dag.substitute_node(new_node, instr, True)
            elif flag == 2:
                dag.remove_op_node(node)
                new_node = dag.substitute_node(new_node, instr, True)
            elif flag == 3:
                dag.remove_op_node(new_node)
                new_node = dag.substitute_node(node, instr, True)

    return dag


def tensor_compiler(qc):
    """
    Compile a quantum circuit for a tensor network simulator,
    i.e. contract all two-qubit gates that are subsequent and
    acts on the same qubits

    Parameters
    ----------
    qc : QuantumCircuit
        quantum circuit to compile

    Return
    ------
    QuantumCircuit
        Compiled quantum circuit
    """
    # Represent the circuit as a directed aciclic graph
    dag = circuit_to_dag(qc)
    # Contract all subsequent 1-qubit gates
    dag = contract_1q_runs(qc, dag)
    # Contract all subsequent 1-qubit and 2-qubits gate
    final_dag = contract_2q_runs(qc, dag)
    # Go back to a quantum circuit
    compiled_qc = dag_to_circuit(final_dag)

    return compiled_qc



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
General utility functions and classes for the simulation
"""

import json

# Import necessary packages
import os
import pickle
import time
from typing import OrderedDict

import numpy as np
import qiskit.qpy as qpy_serialization
from qiskit.circuit import QuantumCircuit
from qtealeaves import write_nml
from qtealeaves.abstracttns.abstract_tn import _AbstractTN
from qtealeaves.convergence_parameters import TNConvergenceParameters
from qtealeaves.observables import TNObservables

from .tn_utils import QCOperators

__all__ = [
    "print_state",
    "fidelity",
    "QCCheckpoints",
    "QCIO",
    "QCConvergenceParameters",
    "SimpleHamiltonian",
]


class SimpleHamiltonian(dict):
    """
    Simple class for an Hamiltonian that extends a normal dictionary.
    The keys are the pauli strings, the values the coefficients.
    It is used for simplicity, since it has a `to_pauli_dict` method
    equivalent to qiskit and other methods to ease the construction.
    """

    def set_num_qubits(self, num_qubits):
        """
        Set the number of qubits the Hamiltonian is describing

        Parameters
        ----------
        num_qubits : int
            Number of qubits
        """
        self["num_qubits"] = num_qubits

    def add_term(self, hterms, qubits, coeff):
        """
        Add a term to the Hamiltonian acting on the
        qubits qubits. You do not need to specify the identities

        Parameters
        ----------
        hterms : str or array-like
            Pauli matrices to apply
        qubits : int or array-like
            Qubits where the terms acts
        coeff : complex
            Coefficient of the term

        Returns
        -------
        None
        """
        if np.isscalar(qubits):
            qubits = np.array([qubits])
            hterms = np.array([hterms])
        ordering = np.argsort(qubits)
        qubits = qubits[ordering]
        hterms = hterms[ordering]

        pauli_string = ""
        for hterm, qubit in zip(hterms, qubits):
            last_qubit = len(pauli_string)
            pauli_string += "I" * (qubit - last_qubit)
            pauli_string += hterm
        last_qubit = len(pauli_string)
        pauli_string += "I" * (self["num_qubits"] - last_qubit)

        self[pauli_string[::-1]] = coeff

    def to_pauli_dict(self):
        """
        Get the qiskit pauli dict representation, that can be later
        used in the observable class

        Returns
        -------
        dict
            dictionary with qiskit pauli_dict old format
        """
        pauli_dict = {"paulis": []}
        for key, val in self.items():
            if key == "num_qubits":
                continue

            pauli_dict["paulis"].append(
                {"label": key, "coeff": {"real": np.real(val), "imag": np.imag(val)}}
            )
        return pauli_dict


class QCCheckpoints:
    """
    Class to handle checkpoint parameters

    Parameters
    ----------
    PATH: str, optional
        PATH to the checkpoint directory. Default `data/checkpoints/`.
    frequency: float, optional
        Decide the frequency, in **hours**, of the checkpoints.
        If negative no checkpoints are present. Default to -1.
    input_nml: str, optional
        Name of the input namelist. Default 'input.nml'
    restart: int, optional
        If an int is provided, it is the checkpoint counter from which the user wants to restart.
        Default to None.
    """

    def __init__(
        self,
        PATH="data/checkpoints/",
        frequency=-1,
        input_nml="input.nml",
        restart=None,
    ):
        # pylint: disable-next=invalid-name
        self._PATH = PATH if (PATH.endswith("/")) else PATH + "/"
        self._frequency = frequency
        self._input_nml = input_nml
        self.restart = restart
        self._checkpoint_cnt = 0
        self._initial_line = 0
        self._initial_time = -1

    def set_up(
        self,
        input_dict,
        operators=QCOperators(),
        observables=TNObservables(),
        circ="",
    ):
        """Set up the checkpoints directory

        Parameters
        ----------
        input_dict : dict
            Input parameter dictionary
        operators : :py:class:`QCOperators`, optional
            Tensor operators
        obervables : :py:class: `TNObservables`, optional
            Tensor observables
        circ_str: str or QuantumCircuit
            String representing the qiskit quantum circuit
        """
        if not isinstance(operators, QCOperators):
            raise TypeError("Operators must be QCOperators type")
        if not isinstance(observables, TNObservables):
            raise TypeError("observables must be TNObservables type")

        if not os.path.isdir(self.PATH):
            os.mkdir(self.PATH)

        # Modify for new PATH
        input_dict["inPATH"] = self.PATH

        # Write files that can be already written
        with open(os.path.join(self.PATH, "observables.pk"), "wb") as fh:
            pickle.dump(observables, fh)
        _, operator_mapping = operators.write_input(self.PATH, {}, 2, {})
        observables.write(self.PATH, {}, operator_mapping)
        write_nml("INPUT", input_dict, os.path.join(self.PATH, self.input_nml))

        # We save extra infos for checkpoints, i.e. the quantum circuit
        if isinstance(circ, str):
            with open(os.path.join(self.PATH, "circuit.dat"), "w") as fh:
                fh.write(circ)
        elif isinstance(circ, QuantumCircuit):
            with open(os.path.join(self.PATH, "circuit.qpy"), "wb") as fd:
                qpy_serialization.dump(circ, fd)
        else:
            raise ValueError(f"Impossible to handle circuit of type {type(circ)}")

        self._initial_time = time.time()

    @property
    def PATH(self):
        """PATH property"""
        return self._PATH

    @property
    def frequency(self):
        """Checkpoint frequency property"""
        return self._frequency

    @property
    def input_nml(self):
        """Input namelist property"""
        return self._input_nml

    def save_checkpoint(self, operation_idx, emulator):
        """
        Save the state for the checkpoint if the
        `operation_idx` exceeded the frequency of the
        checkpoints

        Parameters
        ----------
        operation_idx : int
            Index of the current operation in the quantum circuit
        emulator: _AbstractTN
            Tensor network class

        Returns
        -------
        None
        """
        elapsed_time = (time.time() - self._initial_time) / 3600
        if elapsed_time > self.frequency > 0:
            dir_path = os.path.join(self.PATH, str(self._checkpoint_cnt))

            # Create directory if not accessible
            if not os.path.isdir(dir_path):
                os.makedirs(dir_path)

            # Save the TN state on file.
            file_path = os.path.join(dir_path, "tn_state.npy")
            tensor_list = emulator.to_tensor_list()
            np.save(file_path, np.array(tensor_list, dtype=object), allow_pickle=True)

            # Save the index of the line on file
            # The +1 is because that operation has already been applied
            with open(os.path.join(dir_path, "index.txt"), "w") as fh:
                fh.write(str(operation_idx + 1))

            # Update the counter
            self._checkpoint_cnt += 1

            # Restart the countdown to the next checkpoint
            self._initial_time = time.time()

    def restart_from_checkpoint(self, initial_state):
        """
        Restart from the checkpoint passed in the initialization.

        Parameters
        ----------
        initial_state: str | List[Tensors] | _AbstractTN
            The initial state of the simulation. Might be overwritten on exit.

        Returns
        -------
        str | List[Tensors] | _AbstractTN
            The new initial state. If `self.restart` is None, it is the old
            initial state.
        """
        # Default value, no restart was requested
        if self.restart is None:
            return initial_state

        # If the value is -1, restart from the last one
        if self.restart == -1:
            self.restart = 0
            while os.path.isdir(os.path.join(self.PATH, str(self.restart))):
                self.restart += 1
            self.restart -= 1

        dir_path = os.path.join(self.PATH, str(self.restart))

        # Save the index of the line on file
        with open(os.path.join(dir_path, "index.txt"), "r") as fh:
            self._initial_line = int(fh.read())

        # Read the TN state
        initial_state = np.load(
            os.path.join(dir_path, "tn_state.npy"), allow_pickle=True
        )

        return initial_state

    def to_dict(self):
        """
        Return the ordered dictionary of the properties of the class.

        Returns
        -------
        dictionary: OrderedDict
            Ordered dictionary of the class properties

        """
        dictionary = OrderedDict()

        dictionary["checkpoint_PATH"] = self.PATH
        dictionary["checkpoint_frequency"] = self.frequency
        dictionary["initial_line"] = 1

        return dictionary

    def to_json(self, path):
        """
        Write the class as a json on file
        """
        path = os.path.join(path, "checkpoints.json")
        with open(path, "w") as fp:
            json.dump(self.to_dict(), fp, indent=4)

    @classmethod
    def from_json(cls, path):
        """
        Initialize the class from a json file
        """
        path = os.path.join(path, "checkpoints.json")
        with open(path, "r") as fp:
            dictionary = json.load(fp)[0]

        return cls(**dictionary)


class QCIO:
    """
    Class to handle Input/Output parameters

    Parameters
    ----------
    inPATH: str, optional
        PATH to the directory containing the input files.
        Default to 'data/in/'
    outPATH: str, optional
        PATH to the directory containing the output files.
        Default to 'data/out/'
    initial_state: str or :py:class:`MPS`, optional
        If an MPS, then the list of tensors is used as initial state for
        a starting point of the simulation.
        If 'Vacuum' start from |000...0>. Default to 'Vacuum'.
        If a PATH it is a PATH to a saved MPS.
    """

    initial_states_keywords = ("vacuum", "random")

    def __init__(
        self,
        inPATH="data/in/",
        outPATH="data/out/",
        initial_state="Vacuum",
    ):
        # pylint: disable-next=invalid-name
        self._inPATH = inPATH if inPATH.endswith("/") else inPATH + "/"
        # pylint: disable-next=invalid-name
        self._outPATH = outPATH if outPATH.endswith("/") else outPATH + "/"
        self._initial_state = initial_state

    def setup(self):
        """
        Setup the io files
        """

        # Directories
        if not os.path.isdir(self.inPATH):
            os.makedirs(self.inPATH)
        if not os.path.isdir(self.outPATH):
            os.makedirs(self.outPATH)

        # Initial state

        # First, check if it is a string.
        if isinstance(self.initial_state, str):
            if self.initial_state.lower() not in self.initial_states_keywords:
                # Handle the string case assuming it is a path
                if not os.path.isfile(self.initial_state):
                    raise FileNotFoundError("Path to input file does not exist.")

        else:
            # Assume it is an _AbstractTN that we can write in a formatted way
            if hasattr(self.initial_state, "write"):
                self.initial_state.write(os.path.join(self.inPATH, "initial_state"))
            elif hasattr(self.initial_state, "save_pickle"):
                self.initial_state.save_pickle(
                    os.path.join(self.inPATH, "initial_state")
                )
            # Assume it is a list and use numpy save
            else:
                np.save(
                    os.path.join(self.inPATH, "initial_state"),
                    np.array(self.initial_state, dtype=object),
                    allow_pickle=True,
                )

    @property
    def inPATH(self):
        """Input PATH property"""
        return self._inPATH

    @property
    def outPATH(self):
        """Output PATH property"""
        return self._outPATH

    @property
    def initial_state(self):
        """Initial state property"""
        return self._initial_state

    # @initial_state.setter
    def set_initial_state(self, initial_state):
        """Modify the initial state property"""
        if not isinstance(initial_state, str):
            if not isinstance(initial_state, _AbstractTN):
                raise TypeError(
                    "A non-str initial state must be initialized as an _AbstractTN class"
                )
        self._initial_state = initial_state

    def to_dict(self):
        """
        Return the ordered dictionary of the properties of the class.

        Returns
        -------
        dictionary: OrderedDict
            Ordered dictionary of the class properties
        """
        dictionary = OrderedDict()
        for prop, value in vars(self).items():
            if prop == "_initial_state" and not isinstance(value, str):
                # do not write the tensor to json,
                # just write the type as a placeholder
                value = type(value).__name__
            dictionary[prop[1:]] = value

        return dictionary

    def to_json(self, path):
        """
        Write the class as a json on file
        """
        path = os.path.join(path, "io_info.json")
        with open(path, "w") as fp:
            json.dump(self.to_dict(), fp, indent=4)

    @classmethod
    def from_json(cls, path):
        """
        Initialize the class from a json file
        """
        path = os.path.join(path, "io_info.json")
        with open(path, "r") as fp:
            dictionary = json.load(fp)[0]

        return cls(**dictionary)


class QCConvergenceParameters(TNConvergenceParameters):
    """Convergence parameter class, inhereting from the
    more general Tensor Network type. Here the convergence
    parameters are only the bond dimension and the cut ratio.

    Parameters
    ----------
    max_bond_dimension : int, optional
        Maximum bond dimension of the problem. Default to 10.
    cut_ratio : float, optional
        Cut ratio for singular values. If :math:`\\lambda_n/\\lambda_1 <` cut_ratio then
        :math:`\\lambda_n` is neglected. Default to 1e-9.
    trunc_tracking_mode : str, optional
        Modus for storing truncation, 'M' for maximum, 'C' for
        cumulated (default).
    svd_ctrl : character, optional
        Control for the SVD algorithm. Available:
        - "A" : automatic. Some heuristic is run to choose the best mode for the algorithm.
        The heuristic can be seen in qtealeaves/tensors/tensors.py
        in the function _process_svd_ctrl.
        - "V" : gesvd. Safe but slow method.
        - "D" : gesdd. Fast iterative method. It might fail. Resort to gesvd if it fails
        - "E" : eigenvalue decomposition method. Faster on GPU. Available only when
        contracting the singular value to left or right
        - "X" : sparse eigenvalue decomposition method. Used when you reach the maximum
        bond dimension.
        - "R" : random svd method. Used when you reach the maximum bond dimension.
        Default to 'A'.
    ini_bond_dimension: int, optional
        Initial bond dimension of the simulation. It is used if the initial state is random.
        Default to 1.

    """

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        max_bond_dimension=10,
        cut_ratio=1e-9,
        trunc_tracking_mode="C",
        svd_ctrl="A",
        ini_bond_dimension=1,
    ):
        TNConvergenceParameters.__init__(
            self,
            max_bond_dimension=max_bond_dimension,
            cut_ratio=cut_ratio,
            trunc_tracking_mode=trunc_tracking_mode,
            svd_ctrl=svd_ctrl,
            ini_bond_dimension=ini_bond_dimension,
        )

    def to_dict(self):
        """Return the ordered dictionary of the properties of
        the class

        Returns
        -------
        dictionary: OrderedDict
            Ordered dictionary of the class properties
        """
        dictionary = OrderedDict()
        dictionary["max_bond_dimension"] = self.max_bond_dimension
        dictionary["cut_ratio"] = self.cut_ratio
        dictionary["trunc_tracking_mode"] = self.trunc_tracking_mode

        return dictionary

    def pretty_print(self):
        """
        Print the convergence parameters.
        (Implemented to avoid too few public methods)
        """
        print("-" * 50)
        print(
            "-" * 10 + f" Maximum bond dimension: {self.max_bond_dimension} " + "-" * 10
        )
        print("-" * 10 + f" Cut ratio: {self.cut_ratio} " + "-" * 10)
        print(
            "-" * 10
            + f" Truncation tracking mode: {self.trunc_tracking_mode} "
            + "-" * 10
        )
        print("-" * 50)

    def to_json(self, path):
        """
        Write the class as a json on file
        """
        path = os.path.join(path, "convergence_parameters.json")
        with open(path, "w") as fp:
            dictionary = self.to_dict()
            dictionary["svd_ctrl"] = self.svd_ctrl
            json.dump(dictionary, fp, indent=4)

    @classmethod
    def from_json(cls, path):
        """
        Initialize the class from a json file
        """
        path = os.path.join(path, "convergence_parameters.json")
        with open(path, "r") as fp:
            dictionary = json.load(fp)[0]

        return cls(**dictionary)


def merge_ordered_dicts(dicts):
    """Merge ordered dicts together, concatenating them in the order provided in the list

    Parameters
    ----------
    dicts : list of OrderedDict
        OrderedDict to concatenate

    Return
    ------
    final_dict: OrderedDict
        Concatenated OrderedDict
    """
    for dictionary in dicts:
        if not isinstance(dictionary, OrderedDict):
            raise TypeError("Only OrderedDict can be concatenated using this function")

    final_dict = dicts[0]
    for dictionary in dicts[1:]:
        final_dict.update(dictionary)

    return final_dict


def print_state(dense_state):
    """
    Prints a *dense_state* with kets. Compatible with quimb states.

    Parameters
    ----------
    dense_state: array_like
            Dense representation of a quantum state

    Returns
    -------
    None: None
    """

    nn = int(np.log2(len(dense_state)))

    binaries = [bin(ii)[2:] for ii in range(2**nn)]
    binaries = ["0" * (nn - len(a)) + a for a in binaries]  # Pad with 0s

    ket = []
    for ii, coef in enumerate(dense_state):
        if not np.isclose(np.abs(coef), 0.0):
            if np.isclose(np.imag(coef), 0.0):
                if np.isclose(np.real(coef), 1.0):
                    ket.append("|{}>".format(binaries[ii]))
                else:
                    ket.append("{:.3f}|{}>".format(np.real(coef), binaries[ii]))
            else:
                ket.append("{:.3f}|{}>".format(coef, binaries[ii]))
    print(" + ".join(ket))


def fidelity(psi, phi):
    """
    Returns the fidelity bewteen two quantum states *psi*, *phi* defined as
    :math:`|\\langle\\psi|phi\\rangle|^2`

    Parameters
    ----------
    psi: complex np.array or quimb.core.qarray
            Quantum state
    phi: complex np.array or quimb.core.qarray
            Quantum state

    Returns
    -------
    Fidelity: double real
            Fidelity of the two quantum states
    """

    # Enforcing normalization
    psi /= np.sqrt(np.abs(np.sum(psi.conj() * psi)))
    phi /= np.sqrt(np.abs(np.sum(phi.conj() * phi)))

    return np.abs(np.vdot(psi, phi)) ** 2



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
I/O functions for tensors and mps (list of tensors).
Furthermore, the class :py:class:`QCOperators` for handling operators is provided.

Functions and classes
---------------------

"""

# pylint: disable=too-many-branches

import numpy as np
from qtealeaves import read_tensor
from qtealeaves.operators import TNOperators

__all__ = ["write_tensor", "write_mps", "read_mps", "QCOperators"]


def write_tensor(tensor, dest, cmplx=True, sparse=False):
    """
    Write a tensor stored in a numpy matrix to a file. Conversion
    to column major is taken care of here.

    **Arguments**

    tensor : np.ndarray
        Tensor to be written to the file.

    dest : str, or filehandle
        If string, file will be created or overwritten. If filehandle,
        then data will be written there.

    sparse: bool
        If True write the tensor in a sparse format, i.e. each row is written as
        idx elem
        where idx is the position of the element elem in the tensor vector
    """
    if isinstance(dest, str):
        fh = open(dest, "w+")
    elif hasattr(dest, "write"):
        fh = dest
    else:
        raise TypeError("`dest` for `write_tensor` neither str nor writeable.")

    # Number of links
    fh.write("%d \n" % (len(tensor.shape)))

    # Dimensions of links
    dl = " ".join(list(map(str, tensor.shape)))
    fh.write(dl + "\n")

    # Now we need to transpose
    tensor_colmajor = np.reshape(tensor, -1, order="F")
    if sparse:
        nonzero = np.nonzero(tensor_colmajor)[0]  # index of nonzero element
        tensor_colmajor = tensor_colmajor[nonzero]
        nonzero += 1  # from python to fortran indexing
        fh.write(f"{len(nonzero)} \n")

    for ii, elem in enumerate(tensor_colmajor):
        if cmplx:
            if sparse:
                fh.write(
                    "%d (%30.15E, %30.15E)\n"
                    % (nonzero[ii], np.real(elem), np.imag(elem))
                )
            else:
                fh.write("(%30.15E, %30.15E)\n" % (np.real(elem), np.imag(elem)))
        else:
            if sparse:
                fh.write("%d %30.15E\n" % (nonzero[ii], np.real(elem)))
            else:
                fh.write("%30.15E\n" % (np.real(elem)))
            imag_part = np.imag(elem)
            if np.abs(imag_part) > 1e-14:
                raise TypeError("Writing complex valued tensor as real valued tensor.")

    if isinstance(dest, str):
        fh.close()

    return


def read_mps(filename, cmplx=True):
    """Read an MPS from a given file.
    Reads in column-major order but the output is in row-major.

    This function was used to exchange the MPS with the Fortran
    backend (thus the different ordering), which is now deprecated.

    Parameters
    ----------
    filename: str
            PATH to the file
    cmplx: bool, optional
            If True the MPS is complex, real otherwise. Default to True

    Returns
    -------
    mps: list
            list of np.array in row-major order
    """
    mps = []
    with open(filename, "r") as fh:
        num_sites = int(fh.readline())
        for _ in range(num_sites):
            tens = read_tensor(fh, cmplx=cmplx)
            mps.append(tens)

    return mps


def write_mps(filename, mps, cmplx=True):
    """Write an MPS to file.
    The tensor is written in Fortran-like indexing, i.e. the MPS tensors
    will be converted from row-major into column-major order.

    This function was used to exchange the MPS with the Fortran
    backend (thus the different ordering), which is now deprecated.

    Parameters
    ----------
    filename: str
            PATH to the file
    mps: list
            List of tensors forming the mps
    cmplx: bool, optional
            If True the MPS is complex, real otherwise. Default to True

    Returns
    -------
    None
    """
    with open(filename, "w") as fh:
        fh.write(str(len(mps)) + " \n")
        for tens in mps:
            write_tensor(tens, fh, cmplx=cmplx)

    return None


class QCOperators(TNOperators):
    """
    Class to store and save to file operators. To add an operator to the list use
    self['op_name'] = op_tensor.

    It starts with the pauli operators already defined, with keywords "X","Y","Z"
    """

    def __init__(self):
        TNOperators.__init__(self)
        pauli_matrices = {
            "Z": np.array([[1, 0], [0, -1]]),
            "X": np.array([[0, 1], [1, 0]]),
            "Y": np.array([[0, -1j], [1j, 0]]),
            "I": np.array([[1, 0], [0, 1]]),
        }
        for key, val in pauli_matrices.items():
            self[key] = val



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os
import os.path
import unittest
from io import StringIO

import pylint.lint
from pylint.reporters.text import ParseableTextReporter


class TestPylint(unittest.TestCase):
    """
    Run pylint to check syntax in source files.

    **Details**

    We disable globally:

    * C0325: superfluous parenthesis
    * C0209: consider using fstring
    * W1514: unspecified encoding
    * R1711: useless returns (for allowing empty iterators with
      return-yield)
    * R1732: consider using with
    * Skip Unused argument errors when args
    * Skip Unused argument errors when kargs

    """

    def setUp(self):
        """
        Provide the test setup.
        """
        pattern_1 = (
            "[E0611(no-name-in-module), ] No name '_TNObsBase' in module"
            + " 'qtealeaves'"
        )
        self.default_patterns = [pattern_1]
        self.pylint_args = {
            "good-names": "ii,jj,kk,nn,mm,fh,dx,dy,dz,dt,hh,qc,dl,gh,xp,PATH,inPATH,outPATH",
            "disable": "C0325,C0209,W1514,R1711,R1732",
            "known-third-party": "qtealeaves,qredtea",
            #'ignore_in_line' : pattern_1
        }

    def run_pylint(self, filename, local_settings={}):
        """
        Run linter test with our unit test settings for one specific
        filename.
        """
        args = []

        ignore_in_line = self.default_patterns
        if "ignore_in_line" in local_settings:
            ignore_in_line += local_settings["ignore_in_line"]
            del local_settings["ignore_in_line"]

        for elem in self.pylint_args.keys():
            args += ["--" + elem + "=" + self.pylint_args[elem]]

            if elem in local_settings:
                args[-1] = args[-1] + "," + local_settings[elem]
                del local_settings[elem]

        for elem in local_settings.keys():
            args += ["--" + elem + "=" + local_settings[elem]]

        args += [filename]

        obj = StringIO()
        reporter = pylint.reporters.text.ParseableTextReporter(obj)
        pylint.lint.Run(args, reporter=reporter, exit=False)

        error_list = []
        for elem in obj.getvalue().split("\n"):
            tmp = elem.replace("\n", "")

            if len(tmp) == 0:
                continue
            if tmp.startswith("***"):
                continue
            if tmp.startswith("---"):
                continue
            if tmp.startswith("Your code"):
                continue
            if "Unused argument 'args'" in tmp:
                continue
            if "Unused argument 'kwargs'" in tmp:
                continue

            do_continue = False
            for pattern in ignore_in_line:
                if pattern in tmp:
                    do_continue = True

            if do_continue:
                continue

            error_list.append(tmp)

        return error_list

    def test_folders_recursively(self):
        """
        Recursively run python linter test on all .py files of
        specified folders.
        """
        parent_folders = ["qmatchatea"]
        skip_files = []
        error_list = []

        for elem in parent_folders:
            for root, dirnames, filenames in os.walk(elem):
                for filename in filenames:
                    if not filename.endswith(".py"):
                        continue

                    if filename in skip_files:
                        continue

                    target_file = os.path.join(root, filename)

                    target_attr = "get_settings_" + filename.replace(".py", "")
                    if hasattr(self, target_attr):
                        target_setting = self.__getattribute__(target_attr)()
                    else:
                        target_setting = {}

                    error_list_ii = self.run_pylint(
                        target_file, local_settings=target_setting
                    )

                    error_list += error_list_ii

        if len(error_list) > 0:
            print("\n".join(error_list))

        self.assertEqual(len(error_list), 0, "\n".join(error_list))



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Core function to run the simulation: :py:func:`run_simulation`.

Running the program
-------------------

 It is possible to choose between different approaches for running the simulation:
 - The machine precision can be either double complex "Z" or single complex "C".
 - The device of the simulation can be either "cpu" or "gpu".
 - The number of processes for the MPI simulation. If 1 is passed, then a serial program is run.
 - The approach for the MPI simulation. Either cartesian "CT" or serial "SR".


Serial program
~~~~~~~~~~~~~~

The default way of running this program is to run it serially. While it may be slower than a
parallel implementation it is safer, since the approximation we do on the state is minor.
There are different possibilities to run the program serially:

- Using the CPU backend, which is built in numpy.
- Using the GPU backend (``device='gpu'``).
  The backend is entirely built with cupy,
  runs the simulation on a GPU, giving back the results on the CPU.

Furthermore, there are several options to run the simulation on alternative tensor modules.
Look at the options in :func:`QCBackend <qmatchatea.utils.QCBackend>`.

Parallel program
~~~~~~~~~~~~~~~~

There are three possible algorithm provided to run the program on multiple cores:

- a Cartesian approach (``mpi_approach='CT'``), where the MPS is evenly divided among different
  processes, and each process perform the evolution only on his subsystem
- running multiple independent serial simulations on multiple cores. To know more see
  :doc:`par_simulations`

At the end of the evolution the MPS is brought back together into the master process and the
measurements are performed. The command to call the parallel program should be
:code:`mpirun -np np ./par_main_qmatchatea.exe` where :code:`np` is the
number of processes you will use. It should be given as input to the :class:`IO_INFO`.
For further details on the input class see :doc:`utils`.

Remember that there are communications involved in the MPI parallel program.
You should use it only if your system is large enough or if the bond
dimension used is big enough.

Checking the convergence: Singular values cut
---------------------------------------------

The singular values cut are an indication of the approximation taking place during the
simulation. Indeed, at each application of two-qubit there is an approximation going on,
subject to the *bond_dim* :math:`\\chi` and the *cut_ratio* :math:`\\epsilon` parameters.
Based on the *trunc_tracking_mode* the singular values are reported:

- *trunc_tracking_mode=M*, at each two-site operator we save the maximum of the singular values cut
- *trunc_tracking_mode=C*, at each two-site operator we save the **sum** of the singular values cut

We recall that, given a decreasing-ordered vector of singular values
:math:`S=(s_1, s_2, \\dots, s_n)` we truncate all the singular values:

.. math::

        s_i \\; \\mbox{ is truncated }\\quad \\mbox{if} \\; i>\\chi \\; \\mbox{ or } \\;
        \\frac{s_i}{s_1}<\\epsilon

Observables
-----------

For a detailed description of the observables, i.e. the measurable quantities, please refer
to :doc:`observables`.

"""

import os
import os.path
from typing import OrderedDict

from qtealeaves.observables import TNObservables

from qmatchatea.circuit import Qcircuit
from qmatchatea.py_emulator import run_py_simulation

from .preprocessing import preprocess
from .utils import QCBackend
from .utils.mpi_utils import to_layered_circ
from .utils.qk_utils import qk_transpilation_params
from .utils.tn_utils import QCOperators
from .utils.utils import QCIO, QCCheckpoints, QCConvergenceParameters

__all__ = ["run_simulation"]


# pylint: disable-next=too-many-branches, too-many-arguments
def run_simulation(
    circ,
    local_dim=2,
    convergence_parameters=QCConvergenceParameters(),
    operators=QCOperators(),
    io_info=QCIO(),
    observables=TNObservables(),
    transpilation_parameters=qk_transpilation_params(),
    checkpoints=QCCheckpoints(),
    backend=QCBackend(),
):
    """
    Transpile the circuit to adapt it to the linear structure of the MPS and run the circuit,
    obtaining in output the measurements.

    Parameters
    ----------
    circ: QuantumCircuit | Qcircuit
        Quantum circuit object to simulate.
        Be careful, if one passes a Qcircuit **no check** is done on the quantum circuit, i.e.
        it is not linearized and no barriers are added automatically.
    local_dim: int, optional
        Local dimension of the single degree of freedom. Default is 2, for qubits
    convergence_parameters: :py:class:`QCConvergenceParameters`, optional
        Maximum bond dimension and cut ratio. Default to max_bond_dim=10, cut_ratio=1e-9.
    operators: :py:class:`QCOperators`, optional
        Operator class with the observables operators ALREADY THERE. If None, then it is
        initialized empty. Default to None.
    io_info: :py:class:`QCIO`, optional
        Informations about input/output files.
    observables: :py:class:`TNObservables`, optional
        The observables to be measured at the end of the simulation. Default to TNObservables(),
        which contains no observables to measure.
    transpilation_parameters: :py:class:`qk_transpilation_params`, optional
        Parameters used in the qiskit transpilation phase. Default to qk_transpilation_params().
    checkpoints: :py:class:`QCCheckpoints`, optional
        Class to handle checkpoints in the simulation.
    backend: :py:class:`QCBackend`, optional
        Backend containing all the information for where to run the simulation.

    Returns
    -------
    result: qmatchatea.SimulationResults
        Results of the simulation, containing the following data:
        - Measures
        - Statevector
        - Computational time
        - Singular values cut
        - Entanglement
        - Measure probabilities
        - MPS state
        - Observables measurements
    """
    # Checks on input parameters
    if not isinstance(convergence_parameters, QCConvergenceParameters):
        raise TypeError(
            "convergence_parameters must be of type QCConvergenceParameters"
        )
    if not isinstance(operators, QCOperators):
        raise TypeError("operators must be of type QCOperators")
    if not isinstance(io_info, QCIO):
        raise TypeError("io_info must be of type QCIO")
    if not isinstance(observables, TNObservables):
        raise TypeError("observables must be of type TNObservables")
    if not isinstance(transpilation_parameters, qk_transpilation_params):
        raise TypeError(
            "transpilation_parameters must be of type qk_transpilation_params"
        )
    if not isinstance(checkpoints, QCCheckpoints):
        raise TypeError("checkpoints must be of type QCCheckpoints")
    if not isinstance(backend, QCBackend):
        raise TypeError("backend must be of type QCBackend")

    # Ensure observables output folders is present and set IO
    io_info.setup()

    # Set the PATH to the saved files into the output folder
    for ii in range(len(observables.obs_list["TNState2File"])):
        if "/" not in observables.obs_list["TNState2File"].name[ii]:
            observables.obs_list["TNState2File"].name[ii] = os.path.join(
                io_info.outPATH, observables.obs_list["TNState2File"].name[ii]
            )

    # Preprocess the circuit to adapt it to the MPS constraints (linearity)
    if not isinstance(circ, Qcircuit):
        # Inside the preprocessing, flag the swaps with another name
        # to avoid problems. We don't want the swaps added for keeping
        # the layout linear to be affecting the results of the circuit

        circ = preprocess(circ, qk_params=transpilation_parameters)

        if backend.where_barriers > 0:
            circ = to_layered_circ(circ, where_barriers=backend.where_barriers)

    # Prepare input dictionary
    input_dict = OrderedDict()
    input_dict["num_sites"] = circ.num_qubits
    input_dict["local_dim"] = local_dim
    input_dict["observables_filename"] = os.path.join(observables.filename_observables)

    # Setting checkpoints if required
    if checkpoints.frequency > 0:
        checkpoints.set_up(input_dict, operators, observables, circ)

    # Write all the information of the simulation as json
    to_write = [convergence_parameters, backend, io_info, checkpoints]
    for cls_to_write in to_write:
        cls_to_write.to_json(io_info.inPATH)

    result = run_py_simulation(
        circ,
        local_dim,
        convergence_parameters,
        operators,
        observables,
        backend=backend,
        initial_state=io_info.initial_state,
        checkpoints=checkpoints,
    )

    return result



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os
import os.path
import subprocess
import unittest
from shutil import rmtree

try:
    import mpi4py

    mpi_is_present = True
except ModuleNotFoundError:
    mpi_is_present = False


class TestSimulationParallelization(unittest.TestCase):
    """
    Since it is not possible to directly run this test we perform a workaround:
    through subprocess the unittest calls the actual test
    """

    def setUp(self):
        self.mpi_command = os.environ.get("QMATCHATEA_MPI_COMMAND", "mpiexec")

    def base_mpi_test(self, filename, msg):
        """Run an MPI example as unittest."""
        cmd = [self.mpi_command, "-n", "4", "python3", filename]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            env=os.environ,
            cwd=os.getcwd(),
        )

        success = result.returncode == 0
        if not success:
            print("Was running", " ".join(cmd), "in", os.getcwd())
            print("Stdout", result.stdout)
            print("Stderr", result.stderr)
            print("Return code", result.returncode)

        self.assertTrue(success, msg=msg)

    @unittest.skipIf(not mpi_is_present, "MPI is not present")
    def test_mpi_data_parallelism(self):
        """
        Test if the parallelization of multiple simulations is working using GHZ states.
        """
        filename = "tests/mpi/_testmpi_data_parallelism.py"
        msg = ("Data parallelism simulated parallely on multiple processes correts",)

        self.base_mpi_test(filename, msg)
        return

    @unittest.skipIf(not mpi_is_present, "MPI is not present")
    def test_mpimps_ghz(self):
        """
        Test if the parallelization of multiple simulations is working using GHZ states.
        """
        filename = "tests/mpi/_testmpimps_ghz.py"
        msg = "GHZ circuits simulated parallely on multiple processes correts"

        self.base_mpi_test(filename, msg)
        return

    @unittest.skipIf(not mpi_is_present, "MPI is not present")
    def test_mpimps_qft(self):
        """
        Test if the parallelization of multiple simulations is working using GHZ states.
        """
        filename = "tests/mpi/_testmpimps_qft.py"
        msg = "QFT circuits simulated parallely on multiple processes correts"

        self.base_mpi_test(filename, msg)
        return

    @unittest.skipIf(not mpi_is_present, "MPI is not present")
    def test_example_a(self):
        filename = "examples/mpi/data_parallelism_mpi_example.py"
        msg = "Example data parallelism fails."

        self.base_mpi_test(filename, msg)
        return

    @unittest.skipIf(not mpi_is_present, "MPI is not present")
    def test_example_b(self):
        filename = "examples/mpi/mpi_mps.py"
        msg = "Example MPI MPS fails."

        self.base_mpi_test(filename, msg)
        return



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
In this module we add another level of parallelization on the library.
You can run multiple **independent** simulations in parallel, using the mpi4py library.
To fully understand which are the parameters needed by the function refer to :doc:`interface`.

You should be a little more careful if you want to run simulations this way, since the input
parameters are handled through a list of dictionaries and not singularly. Furthermore,
the way you ran your program is different.
To run a script called `sample.py` with `4` processes you will write on yout shell:

.. code-block::

  mpiexec -n 4 python3 sample.py

.. note::

    Consider that the parallilazion scheme is up to the user, the program will run from the first
    to the last. You should so place first the more computationally demanding simulations to ensure
    an effective parallelization.

"""


import inspect

# Numpy
import os
import os.path
from shutil import rmtree
from typing import OrderedDict

from qmatchatea.utils import QCIO

from .interface import run_simulation
from .utils import SimulationResults

__all__ = ["run_parallel_simulations", "MASTER"]

FREETAG = 0  # Tag to say that the workers are free
WORKTAG = 1  # Tag to send the working parameters to the workers
DIETAG = 2  # Tag to break the MW communications

MASTER = 0


# pylint: disable-next=invalid-name
def run_parallel_simulations(params, MPI):
    """Run multiple simulations in parallel

    Parameters
    ----------
    params : OrderedDict
        Ordered dict where the key is the simulation ID and the value is a dictionary
        with the parameters of the function :py:func:`run_simulation`
    MPI: MPI module
        MPI module to perform the MPI calls

    Returns
    -------
    results : OrderedDict
        Ordered dict where the key is the simulation ID and the value the corresponding
        :py:class:`SimulationResults`

    Raises
    ------
    TypeError
        If the input parameters are not an ordered dict
    """
    if not isinstance(params, OrderedDict):
        raise TypeError("Input parameters must be provided through an OrderedDict")

    # Initialize MPI variables for communications
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()

    if rank == 0:
        if not os.path.isdir("TMP_MPI"):
            os.makedirs("TMP_MPI")
        results = _master(params, MPI)

        if os.path.isdir("TMP_MPI"):
            rmtree("TMP_MPI")
        return results

    _worker(MPI)
    return None


# pylint: disable-next=invalid-name
def _master(params, MPI):
    comm = MPI.COMM_WORLD
    results = OrderedDict()
    requests = []
    error_count = 0
    for sim_id, sim_param in zip(params.keys(), params.values()):
        print(f"**** Done ID={sim_id} *****")
        # Initialize results
        # results[ID] = SimulationResults()

        # Check which process is free
        free_process = comm.recv(buf=None, source=MPI.ANY_SOURCE, tag=FREETAG)

        # Send the data to the worker
        comm.send(obj=sim_param, dest=free_process, tag=WORKTAG)
        comm.send(obj=sim_id, dest=free_process, tag=WORKTAG)

        # Receive the data from the workers with NON-BLOCKING communications
        req = comm.irecv(buf=None, source=free_process, tag=WORKTAG)
        requests.append(req)

    print("----- Finished sending out circuits -----")

    # When you finished wait for all the results
    for sim_id, req in zip(params.keys(), requests):
        path = req.wait()
        if path is None:
            error_count += 1
            results[sim_id] = None
        else:
            results[sim_id] = SimulationResults.read_pickle(path)

    # Send the DIETAG to the workers
    for ii in range(comm.Get_size()):
        if ii == MASTER:
            continue
        comm.send(obj=None, dest=ii, tag=DIETAG)

    # Distribute if any errors occured, raise most recent in each process
    error_count = comm.bcast(error_count, root=MASTER)
    if error_count > 0:
        raise RuntimeError(
            "At least one of the workers failed; "
            "reporting only most recent exception in each worker."
        )

    return results


# pylint: disable-next=invalid-name
def _worker(MPI):
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    comm.send(obj=rank, dest=MASTER, tag=FREETAG)
    status = MPI.Status()

    # Store exception if needed
    exc = None

    while True:
        simulation_parameters = comm.recv(
            buf=None, source=MASTER, tag=MPI.ANY_TAG, status=status
        )
        if status.Get_tag() == DIETAG:
            break

        # Perform the simulation
        sim_id = comm.recv(buf=None, source=MASTER, tag=WORKTAG)
        try:
            simulation_parameters = _check_parameters(simulation_parameters, rank)
            # pylint: disable-next=no-value-for-parameter
            results = run_simulation(*simulation_parameters)

            # Save the results
            path = os.path.join("TMP_MPI", f"{sim_id}.pkl")
            results.save_pickle(path)
            # Send back the results
            comm.send(obj=path, dest=MASTER, tag=WORKTAG)

            # Send the I'm FREE message to the MASTER
            comm.send(obj=rank, dest=MASTER, tag=FREETAG)

        # pylint: disable-next=broad-exception-caught
        except Exception as e:
            # Need to handle any error in the simulation
            exc = e

            # Send back None to signal error to master
            comm.send(None, dest=MASTER, tag=WORKTAG)

            # Send the I'm FREE message to the MASTER
            comm.send(obj=rank, dest=MASTER, tag=FREETAG)

    # Get status if any errors occured, raise most recent in each process
    error_count = 0
    error_count = comm.bcast(error_count, root=MASTER)
    if error_count > 0:
        if exc is None:
            raise RuntimeError("This worker had no problem, but neighboring workers.")
        # Will raise most recent or runtime error no active exception
        raise exc

    return


def _check_parameters(simulation_parameters, rank):
    """Given a dictionary where the keys are the parameter names and the values their values
    returns an ordered tuple of parameters that can directly fit into :py:func:`run_simulation`
    and raises errors if the parameters are not corrected.

    Parameters
    ----------
    simulation_parameters : dict
        Dictionary of the input parameters for the :py:func:`run_simulation` function.
        Notice that the parameters `circ` and `bond_dim` are non-optional, and must so
        be provided, while all the optional parameters can remain non-indicated
    rank : int
        Rank of the process

    Returns
    -------
    params: tuple
        ordered tuple of the parameters for the :py:func:`run_simulation`

    Raises
    ------
    RuntimeError
        If a non-optional parameter of :py:func:`run_simulation` is not provided
    ValueError
        If the chosen parallel approach is not serial
    """

    # Setting for the current simulation
    keys = simulation_parameters.keys()
    params = []

    # Parse default arguments
    sim_info = inspect.getfullargspec(run_simulation)
    arg_names = sim_info.args
    default_values = sim_info.defaults
    num_non_optional_args = len(arg_names) - len(default_values)

    for ii, name in enumerate(arg_names):
        if name in keys:
            # Parameter provided in input_simulation
            value = simulation_parameters[name]
        elif ii < num_non_optional_args:
            # Parameter not provided, but it was non-optional
            raise RuntimeError(
                f"Parameter {name} is a non-optional parameter and must \
                be provided"
            )
        else:
            # Parameter not provided, set default
            value = default_values[ii - num_non_optional_args]

        # Raise exception if the chosen parallel approach is not 'SR', since not implemented yet
        if name == "approach" and value != "SR":
            raise ValueError(
                "Only the serial par_approach can be used when parallelizing \
                over different simulations"
            )

        # Special case for the input/output PATHs if they are not provided, to skip problems
        # of reading/writing from the same file from multiple processes
        if (name == "io_info") and (name not in keys):
            value = QCIO(f"data/in/rank_{rank}", f"data/out/rank_{rank}")

        params.append(value)

    assert len(params) == len(
        arg_names
    ), "Number of parameter not equal to number of arguments"
    params = tuple(params)
    return params



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Class to handle the intermediate measurement of observables.

It also handles the indexing of quantum registers and observables.
"""
from copy import deepcopy

import numpy as np
from qtealeaves.observables import TNObservables

from qmatchatea.utils.tn_utils import QCOperators

from .operations import ClassicalCondition, QCOperation


class QCObservableStep(QCOperation):
    """
    Perform a measurement of the observables TNObservables.
    The observables are applied on the **WHOLE** circuit,
    regardless of the register. Thus, this class also
    takes a snapshot of the current quantum registers,
    in such a way to be able to recover the desired
    indexes.

    Parameters
    ----------
    name : str
        Name of the observables step
    observables: :class:`TNObservables`
        Observables to be measured at this step
    operators: :class:`QCOperators`, optional
        Class containing the operators present in the Observables
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.

    Examples
    --------

    For a detailed explanation of how the observable structure will
    change adapting to the qregisters see:

    - `LocalObs`_.
    - `ProjectiveObs`_.
    - `ProbabilityObs`_.
    """

    def __init__(
        self,
        name,
        observables,
        operators=QCOperators(),
        conditioned=ClassicalCondition(),
    ):
        if isinstance(observables, TNObservables):
            self.observables = deepcopy(observables)
            self.operators = operators
            self.qregisters = np.repeat(None, 2)
            self.num_sites = 0
            super().__init__(name, None, None, conditioned)
        else:
            raise TypeError(
                f"observables is type {type(observables)}, not TNObservables"
            )

    @property
    def has_single_qreg(self):
        """True if the observable is specific to a single register, False otherwise"""
        return len(self.qregisters) == 1

    def snap_current_qregisters(self, qregisters):
        """
        Save the current quantum registers in the quantum circuit
        to recover the correct pattern of observable-register
        afterwards

        Parameters
        ----------
        qregisters : dict
            Dictionaty of the quantum registers
        """
        self.qregisters = qregisters

    def adjust_obs_indexing(self):
        """
        Modify the indexing of the observables
        that targets a specific site index, such
        as the TNObsTensorProduct and TNObsWeightedSum.
        """

        # Adjust TP observables
        tensor_prod = self.observables.obs_list["TNObsTensorProduct"]
        if len(tensor_prod) > 0:
            self.observables.obs_list["TNObsTensorProduct"] = self._adjust_tp_indexing(
                tensor_prod
            )

        # Adjust weighted sum observables
        weight_s = self.observables.obs_list["TNObsWeightedSum"]
        if len(weight_s) > 0:
            new_tp_ops = []
            for tp_obs in weight_s.tp_operators:
                new_tp_ops.append(self._adjust_tp_indexing(tp_obs))
            weight_s.tp_operators = new_tp_ops
            self.observables.obs_list["TNObsWeightedSum"] = weight_s

    def postprocess_obs_indexing(self):
        """
        Postprocess the indexing of the observables, after the computation,
        due to the quantum registers.
        """
        # Adjust local observables
        local = self.observables.obs_list["TNObsLocal"]
        if len(local) > 0:
            local = self._postprocess_local(local)
            self.observables.obs_list["TNObsLocal"] = local

        # Adjust projective observables
        proj = self.observables.obs_list["TNObsProjective"]
        if proj.num_shots > 0:
            proj = self._postprocess_projective(proj)
            self.observables.obs_list["TNObsProjective"] = proj

        # Adjust probabilities observables
        prob = self.observables.obs_list["TNObsProbabilities"]
        if len(prob) > 0:
            prob = self._postprocess_probabilities(prob)
            self.observables.obs_list["TNObsProbabilities"] = prob

    def _adjust_tp_indexing(self, tp_obs):
        """
        Adjust the indexing of a TNObsTensorProduct observable.
        This procedure works only if we have a single register
        in the snapshot.

        Parameters
        ----------
        tp_obs : :class:`TNObsTensorProduct`
            Tensor product observable

        Returns
        -------
        :class:`TNObsTensorProduct`
            The new observable with the correct indexing
        """
        if not self.has_single_qreg:
            raise RuntimeError(
                "Cannot adjust indexing when more than" + " a register is involved"
            )

        new_indexing = list(self.qregisters.values())[0]

        new_sites = []
        for idx, obs in enumerate(tp_obs.sites):
            new_sites.append([])
            for site in obs:
                new_sites[idx].append(new_indexing[site])
        tp_obs.sites = new_sites

        return tp_obs

    def _postprocess_local(self, local):
        """
        Postprocess local observables. In input, you get a
        vector of local observables, one for each site.
        In output, the vector of local observable is transformed
        into a dictionary, where the key is the quantum register
        and the value the local measurement corresponding to
        that register

        .. _LocalObs:

        Example
        -------
        Let us suppose that we have a chain of 4 qubits, and we
        want to measure :math:`\\sigma_z` locally. Qubits 0,1 are
        part of register ``'default'``, while qubits 2,3 of
        ``'ancilla'``. Then:

        .. code-block:: python

            # local is the local observable class
            local_obs = local.results_buffer['sz']
            print(local_obs)
            >>>> [0, 1, -1, 0]
            # qcobs is the parent class
            postprocessed_local_obs = qcobs._postprocess_local(local)
            post_proc_res = postprocessed_local_obs.results_buffer['sz']
            print(post_proc_res)
            >>>> {'default':[0, 1], 'ancilla':[-1, 0]}

        Parameters
        ----------
        local : :class:`TNObsLocal`
            Local observables to be postprocessed

        Returns
        -------
        :class:`TNObsLocal`
            Postprocessed local observables
        """

        new_results = {}
        for key, obs in local.results_buffer.items():
            new_results[key] = {}
            for qreg_key, indexing in self.qregisters.items():
                new_results[key][qreg_key] = obs[indexing]
        local.results_buffer = new_results

        return local

    def _postprocess_projective(self, projective):
        """
        Postprocess projective measurements. In input,
        you get the dictionary of the measurement,
        in output a dictionary further divided by the
        quantum registers, as shown in the example.

        .. _ProjectiveObs:

        Example
        -------
        Let us suppose that we have a chain of 4 qubits.
        Qubits 0,1,2 are part of register `'default'`
        and are in a GHZ state, while qubits 3 of
        `'ancilla'` and is in the |1> state. Then:

        .. code-block:: python

            # proj is the projective measuremenent observable class
            proj_obs = proj.results_buffer['projective_measurements']
            print(proj_obs)
            >>>> {'1111' : 512, '1000' : 512}
            # qcobs is the parent class
            pp_proj = qcobs._postprocess_projective(proj)
            post_proc_res = pp_proj.results_buffer['projective_measurements']
            print(post_proc_res)
            >>>> {'default':{'000':512, '111':512}, 'ancilla':{'1':1024}}

        Parameters
        ----------
        local : :class:`TNObsProjective`
            Projective measurement observables to be postprocessed

        Returns
        -------
        :class:`TNObsProjective`
            Postprocessed projective measurement observables
        """

        measures = projective.results_buffer["projective_measurements"]

        new_measures = {}
        for qreg_key, indexing in self.qregisters.items():
            new_measures[qreg_key] = {}
            indexing = self.num_sites - 1 - indexing

            for key, value in measures.items():
                if "," in key:
                    key = np.array(key.split(","))
                    sep = ","
                else:
                    key = np.array(list(key))
                    sep = ""
                key = np.array2string(key[indexing].astype(int), separator=sep)
                key = key.replace("[", "").replace("]", "")

                if key in new_measures[qreg_key]:
                    new_measures[qreg_key][key] += value
                else:
                    new_measures[qreg_key][key] = value
        projective.results_buffer["projective_measurements"] = new_measures

        return projective

    def _postprocess_probabilities(self, probabilities):
        """
        Postprocess probability measurements. In input,
        you get the dictionary of the probabilities,
        in output a dictionary further divided by the
        quantum registers, as shown in the example.
        While probabilities are summed for the even
        and greedy case, when focusing on a quantum register,
        they are instead concatenated in the unbiased case,
        since in that scenario we measure the probability intervals

        .. _ProbabilityObs:

        Examples
        --------
        Let us suppose that we have a chain of 4 qubits.
        Qubits 0,1,2 are part of register `'default'`
        and are in a GHZ state, while qubits 3 of
        `'ancilla'` and is in the |1> state.

        **Even or greedy probabilities**

        Probabilities on the same state are summed.

        .. code-block:: python

            # prob is the probability measuremenent observable class
            prob_obs = prob.results_buffer['even_probabilities']
            print(prob_obs)
            >>>> {'1111' : 0.5, '1000' : 0.5}
            # qcobs is the parent class
            pp_prob = qcobs._postprocess_probabilities(prob)
            post_proc_res = pp_prob.results_buffer['even_probabilities']
            print(post_proc_res)
            >>>> {'default':{'000':0.5, '111':0.5}, 'ancilla':{'1':1}}

        **Unbiased probabilities**

        In this specific case the intervals could be summed, since they
        are contigous, but in the more general case they simply
        cannot, without losing informations.

        .. code-block:: python

            # prob is the probability measuremenent observable class
            prob_obs = prob.results_buffer['unbiased_probabilities']
            print(prob_obs)
            >>>> {'1000' : (0, 0.5), '1111' : (0.5, 1)}
            # qcobs is the parent class
            pp_prob = qcobs._postprocess_probabilities(prob)
            post_proc_res = pp_prob.results_buffer['unbiased_probabilities']
            print(post_proc_res)
            >>>> {'default':{'000':(0, 0.5), '111':(0.5, 1)},
            >>>>    'ancilla':{'1':(0, 0.5, 0.5, 1) }}

        Parameters
        ----------
        probabilities : :class:`TNObsProbabilities`
            Probabilities measurement observables to be postprocessed

        Returns
        -------
        :class:`TNObsProbabilities`
            Postprocessed probabilities measurement observables
        """

        new_probs = {}

        for prob_key, prob in probabilities.results_buffer.items():
            new_probs[prob_key] = {}
            for qreg_key, indexing in self.qregisters.items():
                new_probs[prob_key][qreg_key] = {}
                indexing = self.num_sites - 1 - indexing
                if isinstance(prob, tuple):
                    # in case qtealeaves also return the samples
                    # (see `do_return_samples=True` option)
                    prob = prob[0]
                for key, value in prob.items():
                    if "," in key:
                        key = np.array(key.split(","))
                        sep = ","
                    else:
                        key = np.array(list(key))
                        sep = ""

                    key = np.array2string(key[indexing].astype(int), separator=sep)
                    key = key.replace("[", "").replace("]", "")

                    if key in new_probs[prob_key][qreg_key]:
                        new_probs[prob_key][qreg_key][key] += value
                    else:
                        new_probs[prob_key][qreg_key][key] = value

        probabilities.results_buffer = new_probs

        return probabilities
